# These are experimental IPython magics, providing quick shortcuts for simple
# tasks. None of these save any data.

# To use, run this in an IPython shell:
# ip = get_ipython()
# ip.register_magics(BlueskyMagics)

import asyncio
import collections
import warnings
from operator import attrgetter

import numpy as np
from IPython.core.magic import Magics, line_magic, magics_class
from traitlets import MetaHasTraits

from bluesky import RunEngine, RunEngineInterrupted
from bluesky.protocols import Movable, Readable
from bluesky.utils import ProgressBarManager

from . import plan_stubs as bps
from . import plans as bp

try:
    # cytools is a drop-in replacement for toolz, implemented in Cython
    from cytoolz import partition
except ImportError:
    from toolz import partition


# This is temporarily here to allow for warnings to be printed
# we changed positioners to a property but since we never instantiate
# the class we need to add this
class MetaclassForClassProperties(MetaHasTraits, type):
    @property
    def positioners(self):
        if self._positioners:
            warnings.warn(
                "BlueskyMagics.positioners is deprecated. Please use the newer labels feature.", stacklevel=1
            )
        return self._positioners

    @positioners.setter
    def positioners(self, val):
        warnings.warn(
            "BlueskyMagics.positioners is deprecated. Please use the newer labels feature.", stacklevel=1
        )
        self._positioners = val

    @property
    def detectors(self):
        if self._detectors:
            warnings.warn(
                "BlueskyMagics.detectors is deprecated. Please use the newer labels feature.", stacklevel=1
            )
        return self._detectors

    @detectors.setter
    def detectors(self, val):
        warnings.warn("BlueskyMagics.detectors is deprecated. Please use the newer labels feature.", stacklevel=1)
        self._detectors = val

    _positioners: list[Movable] = []
    _detectors: list[Readable] = []


@magics_class
class BlueskyMagics(Magics, metaclass=MetaclassForClassProperties):
    """
    IPython magics for bluesky.

    To install:

    >>> ip = get_ipython()
    >>> ip.register_magics(BlueskyMagics)

    Optionally configure default detectors and positioners by setting
    the class attributes:

    * ``BlueskyMagics.detectors``
    * ``BlueskyMagics.positioners``

    For more advanced configuration, access the magic's RunEngine instance and
    ProgressBarManager instance:

    * ``BlueskyMagics.RE``
    * ``BlueskyMagics.pbar_manager``
    """

    RE = RunEngine({}, loop=asyncio.new_event_loop())
    pbar_manager = ProgressBarManager()

    def _ensure_idle(self):
        if self.RE.state != "idle":
            print("The RunEngine invoked by magics cannot be resumed.")
            print("Aborting...")
            self.RE.abort()

    @line_magic
    def mov(self, line):
        if len(line.split()) % 2 != 0:
            raise TypeError("Wrong parameters. Expected: %mov motor position (or several pairs like that)")
        args = []
        for motor, pos in partition(2, line.split()):
            args.append(eval(motor, self.shell.user_ns))
            args.append(eval(pos, self.shell.user_ns))
        plan = bps.mv(*args)
        self.RE.waiting_hook = self.pbar_manager
        try:
            self.RE(plan)
        except RunEngineInterrupted:
            ...
        self.RE.waiting_hook = None
        self._ensure_idle()
        return None

    @line_magic
    def movr(self, line):
        if len(line.split()) % 2 != 0:
            raise TypeError("Wrong parameters. Expected: %mov motor position (or several pairs like that)")
        args = []
        for motor, pos in partition(2, line.split()):
            args.append(eval(motor, self.shell.user_ns))
            args.append(eval(pos, self.shell.user_ns))
        plan = bps.mvr(*args)
        self.RE.waiting_hook = self.pbar_manager
        try:
            self.RE(plan)
        except RunEngineInterrupted:
            ...
        self.RE.waiting_hook = None
        self._ensure_idle()
        return None

    @line_magic
    def ct(self, line):
        # If the deprecated BlueskyMagics.detectors list is non-empty, it has
        # been configured by the user, and we must revert to the old behavior.
        if type(self).detectors:
            if line.strip():
                dets = eval(line, self.shell.user_ns)
            else:
                dets = type(self).detectors
        else:
            # new behaviour
            devices_dict = get_labeled_devices(user_ns=self.shell.user_ns)
            if line.strip():
                if "[" in line or "]" in line:
                    raise ValueError(
                        "It looks like you entered a list like "
                        "`%ct [motors, detectors]` "
                        "Magics work a bit differently than "
                        "normal Python. Enter "
                        "*space-separated* labels like "
                        "`%ct motors detectors`."
                    )
                # User has provided a white list of labels like
                # %ct label1 label2
                labels = line.strip().split()
            else:
                labels = ["detectors"]
            dets = []
            for label in labels:
                dets.extend(obj for _, obj in devices_dict.get(label, []))
        plan = bp.count(dets)
        print("[This data will not be saved. Use the RunEngine to collect data.]")
        try:
            self.RE(plan, _ct_callback)
        except RunEngineInterrupted:
            ...
        self._ensure_idle()
        return None

    FMT_PREC = 6

    @line_magic
    def wa(self, line):
        "List positioner info. 'wa' stands for 'where all'."
        # If the deprecated BlueskyMagics.positioners list is non-empty, it has
        # been configured by the user, and we must revert to the old behavior.
        if type(self).positioners:
            if line.strip():
                positioners = eval(line, self.shell.user_ns)
            else:
                positioners = type(self).positioners
            if len(positioners) > 0:
                _print_positioners(positioners, precision=self.FMT_PREC)
        else:
            # new behaviour
            devices_dict = get_labeled_devices(user_ns=self.shell.user_ns)
            if line.strip():
                if "[" in line or "]" in line:
                    raise ValueError(
                        "It looks like you entered a list like "
                        "`%wa [motors, detectors]` "
                        "Magics work a bit differently than "
                        "normal Python. Enter "
                        "*space-separated* labels like "
                        "`%wa motors detectors`."
                    )
                # User has provided a white list of labels like
                # %wa label1 label2
                labels = line.strip().split()
            else:
                # Show all labels.
                labels = list(devices_dict.keys())
            for label in labels:
                print(label)
                try:
                    devices = devices_dict[label]
                    all_children = [
                        (k, getattr(obj, k)) for _, obj in devices for k in getattr(obj, "read_attrs", [])
                    ]
                except KeyError:
                    print("<no matches for this label>")
                    continue
                # Search devices and all their children for positioners.
                positioners = [dev for _, dev in devices + all_children if is_positioner(dev)]
                if positioners:
                    _print_positioners(positioners, precision=self.FMT_PREC, prefix=" " * 2)
                    print()  # blank line
                # Just display the top-level devices in the namespace (no
                # children).
                _print_devices(devices, prefix=" " * 2)
                print()  # blank line


def _print_devices(devices, prefix=""):
    cols = ["Local variable name", "Ophyd name (to be recorded as metadata)"]
    print(prefix + "{:38s} {:38s}".format(*cols))
    for name, obj in devices:
        print(prefix + f"{name:38} {str(obj.name):38s}")


def is_positioner(dev):
    return hasattr(dev, "position")


def _print_positioners(positioners, sort=True, precision=6, prefix=""):
    """
    This will take a list of positioners and try to print them.

    Parameters
    ----------
    positioners : list
        list of positioners

    sort : bool, optional
        whether or not to sort the list

    precision: int, optional
        The precision to use for numbers
    """
    # sort first
    if sort:
        positioners = sorted(set(positioners), key=attrgetter("name"))

    values = []
    for p in positioners:
        try:
            values.append(p.position)
        except Exception as exc:
            values.append(exc)

    headers = ["Positioner", "Value", "Low Limit", "High Limit", "Offset"]
    LINE_FMT = prefix + "{: <30} {: <11} {: <11} {: <11} {: <11}"
    lines = []
    lines.append(LINE_FMT.format(*headers))
    for p, v in zip(positioners, values):
        if not isinstance(v, Exception):
            try:
                prec = int(p.precision)
            except Exception:
                prec = precision
            value = np.round(v, decimals=prec)
            try:
                low_limit, high_limit = p.limits
            except Exception as exc:
                low_limit = high_limit = exc.__class__.__name__
            else:
                low_limit = np.round(low_limit, decimals=prec)
                high_limit = np.round(high_limit, decimals=prec)
            try:
                offset = p.user_offset.get()
            except Exception as exc:
                offset = exc.__class__.__name__
            else:
                offset = np.round(offset, decimals=prec)
        else:
            value = v.__class__.__name__  # e.g. 'DisconnectedError'
            low_limit = high_limit = offset = ""

        lines.append(LINE_FMT.format(p.name, value, low_limit, high_limit, offset))
    print("\n".join(lines))


def get_labeled_devices(user_ns=None, maxdepth=6):
    """Returns dict of labels mapped to devices with that label

    Parameters
    ----------
    user_ns : dict, optional
        The namespace to search on
        Default is to grab the namespace of the ipython shell.

    maxdepth: int, optional
        max recursion depth

    Returns
    -------
        A dictionary of (name, ophydobject) tuple indexed by device label.

    Examples
    --------
    Read devices labeled as motors:
        objs = get_labeled_devices()
        my_motors = objs['motors']
    """
    # could be set but lists are more common for users
    obj_list = collections.defaultdict(list)

    if maxdepth <= 0:
        warnings.warn("Recursion limit exceeded. Results will be truncated.")  # noqa: B028
        return obj_list

    if user_ns is None:
        from IPython import get_ipython

        user_ns = get_ipython().user_ns

    for key, obj in user_ns.items():
        # ignore objects beginning with "_"
        # (mainly for ipython stored objs from command line
        # return of commands)
        # also check its a subclass of desired classes
        if not key.startswith("_"):
            if hasattr(obj, "_ophyd_labels_"):
                # don't inherit parent labels
                labels = obj._ophyd_labels_
                for label in labels:
                    obj_list[label].append((key, obj))

                if is_parent(obj):
                    # Get direct children (not grandchildren).
                    children = {
                        k: getattr(obj, k)
                        # obj might be a Signal (no read_attrs).
                        for k in getattr(obj, "read_attrs", [])
                        if "." not in k
                    }
                    # Recurse over all children.
                    for c_key, v in get_labeled_devices(user_ns=children, maxdepth=maxdepth - 1).items():
                        items = [(".".join([key, ot[0]]), ot[1]) for ot in v]
                        obj_list[c_key].extend(items)

    # Convert from defaultdict to normal dict before returning.
    return {k: sorted(v) for k, v in obj_list.items()}


def is_parent(dev):
    # return whether a node is a parent
    # should not have component_names, or if yes, should be empty
    # read_attrs needed to check it's an instance and not class itself
    return not isinstance(dev, type) and getattr(dev, "component_names", [])


def _ct_callback(name, doc):
    if name != "event":
        return
    for k, v in doc["data"].items():
        print(f"{k: <30} {v}")



import functools
from typing import Callable, cast

from opentelemetry.trace import Tracer, get_tracer

from .protocols import P
from .utils import MsgGenerator

tracer = get_tracer(__name__)


def trace_plan(tracer: Tracer, span_name: str) -> Callable[[Callable[P, MsgGenerator]], Callable[P, MsgGenerator]]:
    """Wraps a generator function in tracer.start_as_current_span(span_name)"""

    def wrap(f: Callable[P, MsgGenerator]) -> Callable[P, MsgGenerator]:
        @functools.wraps(f)
        def wrap_f(*args: P.args, **kwargs: P.kwargs):
            with tracer.start_as_current_span(span_name):
                yield from f(*args, **kwargs)

        return cast(Callable[P, MsgGenerator], wrap_f)

    return wrap



import itertools
from time import time
from unittest.mock import ANY

import pytest

from bluesky import Msg, RunEngineInterrupted
from bluesky.plan_stubs import declare_stream, trigger_and_read
from bluesky.preprocessors import run_decorator
from bluesky.protocols import Callback, Descriptor, Reading
from bluesky.tests import ophyd, requires_ophyd
from bluesky.utils import ancestry, separate_devices, share_ancestor

from .utils import DocCollector

if ophyd:
    from ophyd import Component as Cpt
    from ophyd import Device, Signal

    class A(Device):
        s1 = Cpt(Signal, value=0)
        s2 = Cpt(Signal, value=0)

    class B(Device):
        a1 = Cpt(A, "")
        a2 = Cpt(A, "")

    class DCM(Device):
        th = Cpt(Signal, value=0)
        x = Cpt(Signal, value=0)


class SigNew:
    def __init__(self, name: str) -> None:
        self.name = name
        self._callbacks: list[Callback] = []

    def read(self) -> dict[str, Reading]:
        return {self.name: dict(value=0, timestamp=0)}  # noqa: C408

    def describe(self) -> dict[str, Descriptor]:
        return {self.name: dict(source="", dtype="number", shape=[])}  # noqa: C408

    def subscribe(self, function: Callback) -> None:
        self._callbacks.append(function)

    def clear_sub(self, function: Callback) -> None:
        self._callbacks.remove(function)

    def _run_subs(self, *args, **kwargs):
        for cb in self._callbacks:
            cb({self.name: dict(value=0, timestamp=time())})  # noqa: C408


@requires_ophyd
def test_ancestry():
    b = B("", name="b")
    assert ancestry(b) == [b]
    assert ancestry(b.a1) == [b.a1, b]
    assert ancestry(b.a1.s1) == [b.a1.s1, b.a1, b]


@requires_ophyd
def test_share_ancestor():
    b1 = B("", name="b1")
    b2 = B("", name="b2")
    assert share_ancestor(b1, b1)
    assert share_ancestor(b1, b1.a1)
    assert share_ancestor(b1, b1.a1.s1)
    assert not share_ancestor(b2, b1)
    assert not share_ancestor(b2, b1.a1)
    assert not share_ancestor(b2, b1.a1.s1)


@requires_ophyd
def test_separate_devices():
    b1 = B("", name="b1")
    b2 = B("", name="b2")
    a = A("", name="a")
    assert separate_devices([b1, b1.a1]) == [b1]
    assert separate_devices([b1.a1, b1.a2]) == [b1.a1, b1.a2]
    assert separate_devices([b1, b2.a1]) == [b1, b2.a1]
    assert separate_devices([b1.a1, b2.a2]) == [b1.a1, b2.a2]
    assert separate_devices([b1, a]) == [b1, a]


@pytest.mark.parametrize("ophyd", ["v1", "v2"])
@requires_ophyd
def test_monitor(RE, ophyd):
    docs = []

    def collect(name, doc):
        docs.append(doc)

    if ophyd == "v1":
        sig = A("", name="a").s1
    else:
        sig = SigNew(name="a_s1")

    def plan():
        yield Msg("open_run")
        yield Msg("monitor", sig)
        sig._run_subs(sub_type="value")
        yield Msg("close_run")

    RE(plan(), collect)
    assert len(docs) == 4
    assert docs[2] == {
        "data": {"a_s1": 0},
        "descriptor": ANY,
        "seq_num": 1,
        "time": pytest.approx(time(), rel=0.1),
        "timestamps": {"a_s1": pytest.approx(time(), rel=0.1)},
        "filled": ANY,
        "uid": ANY,
    }


@requires_ophyd
def test_monitor_with_pause_resume(RE):
    docs = []

    def collect(name, doc):
        docs.append(doc)

    a = A("", name="a")

    def plan():
        yield Msg("open_run")
        yield Msg("monitor", a.s1)
        yield Msg("checkpoint")
        a.s1._run_subs(sub_type="value")
        yield Msg("pause")
        a.s1._run_subs(sub_type="value")
        yield Msg("close_run")

    with pytest.raises(RunEngineInterrupted):
        RE(plan(), collect)
    assert len(docs) == 3  # RunStart, EventDescriptor, one Event
    # All but one of these will be ignored. Why is one not ignored, you ask?
    # Beacuse ophyd runs subscriptions when they are (re-)subscriped.
    a.s1._run_subs(sub_type="value")
    a.s1._run_subs(sub_type="value")
    a.s1._run_subs(sub_type="value")
    a.s1._run_subs(sub_type="value")
    a.s1._run_subs(sub_type="value")
    a.s1._run_subs(sub_type="value")
    assert len(docs) == 3
    RE.resume()
    assert len(docs) == 6  # two new Events + RunStop


def _make_overlapping_raising_tests(func):
    labels = ["part_v_whole", "whole_v_part", "different_obj_same_names", "different_names"]
    if ophyd:
        dcm = DCM("", name="dcm")
        dcm2 = DCM("", name="dcm")
        dcm3 = DCM("", name="dcm3")
        dets = ((dcm.th, dcm), (dcm, dcm.th), (dcm, dcm2), (dcm, dcm3))
    else:
        dets = ((None, None),) * len(labels)

    return pytest.mark.parametrize("det1,det2", dets, ids=labels)(func)


@requires_ophyd
@_make_overlapping_raising_tests
def test_overlapping_raise(RE, det1, det2):
    @run_decorator()
    def test_plan(det1, det2):
        yield from declare_stream(det1, name="primary")
        yield from trigger_and_read([det1])
        yield from trigger_and_read([det2])

    with pytest.raises(RuntimeError):
        RE(test_plan(det1, det2))


def _make_overlapping_tests_2stream(func):
    labels = ["part_v_whole", "whole_v_part", "different", "same"]
    if ophyd:
        dcm = DCM("", name="dcm")
        dcm2 = DCM("", name="dcm")
        dets = ((dcm.th, dcm), (dcm, dcm.th), (dcm, dcm2), (dcm, dcm))
    else:
        dets = ((None, None),) * len(labels)

    return pytest.mark.parametrize("det1,det2", dets, ids=labels)(func)


@requires_ophyd
@_make_overlapping_tests_2stream
def test_keyoverlap_2stream(RE, det1, det2):
    @run_decorator()
    def test_plan(det1, det2):
        yield from declare_stream(det1, name="primary")
        yield from trigger_and_read([det1])
        yield from declare_stream(det2, name="other")
        yield from trigger_and_read([det2], name="other")

    d = DocCollector()
    rs = RE(test_plan(det1, det2), d.insert)
    if RE.call_returns_result:
        uid = rs.run_start_uids[0]
    else:
        uid = rs[0]

    assert len(d.start) == 1
    assert len(d.descriptor[uid]) == 2


def _make_overlapping_tests_stream(func):
    labels = ["whole_whole", "whole_v_w+th", "whole_v_w+x", "whole_v_w+x+th"]
    if ophyd:
        dcm = DCM("", name="dcm")
        dets = ((dcm, [dcm]), (dcm, [dcm, dcm.th]), (dcm, [dcm, dcm.x]), (dcm, [dcm, dcm.x, dcm.th]))
    else:
        dets = ((None, None),) * len(labels)

    return pytest.mark.parametrize("det1,det_list", dets, ids=labels)(func)


@requires_ophyd
@_make_overlapping_tests_stream
def test_overlapped_but_identical(RE, det1, det_list):
    @run_decorator()
    def test_plan(det1, det_list):
        yield from declare_stream(det1, name="primary")
        yield from trigger_and_read([det1])
        for p in itertools.permutations(det_list):  # noqa: B007
            yield from trigger_and_read(det_list)

    d = DocCollector()
    rs = RE(test_plan(det1, det_list), d.insert)
    if RE.call_returns_result:
        uid = rs.run_start_uids[0]
    else:
        uid = rs[0]

    assert len(d.start) == 1
    assert len(d.descriptor[uid]) == 1


@requires_ophyd
def test_read_clash(RE):
    dcm = DCM("", name="dcm")
    dcm2 = DCM("", name="dcm")

    with pytest.raises(ValueError):
        RE(
            [
                Msg("open_run"),
                Msg("declare_stream", None, dcm, dcm2.th, name="primary"),
                *trigger_and_read([dcm, dcm2.th]),
                Msg("close_run"),
            ]
        )

    with pytest.raises(ValueError):
        RE(
            [
                Msg("open_run"),
                Msg("declare_stream", None, dcm, dcm2, name="primary"),
                *trigger_and_read([dcm, dcm2]),
                Msg("close_run"),
            ]
        )

    with pytest.raises(ValueError):
        RE(
            [
                Msg("open_run"),
                Msg("declare_stream", None, dcm.th, dcm2.th, name="primary"),
                *trigger_and_read([dcm.th, dcm2.th]),
                Msg("close_run"),
            ]
        )



import logging

from bluesky.log import validate_level


def make_record(level, doc_name):
    levelno = validate_level(level)
    record = logging.getLogRecordFactory()("", levelno, "", "", "", "", "")
    if doc_name is not None:
        record.doc_name = doc_name
    return record


def test_noninterfering_handlers():
    class TestHandler(logging.Handler):
        "Collects all the records it emits in a list for checking in test."

        def __init__(self):
            self.records = []
            super().__init__()

        def emit(self, record):
            self.records.append(record)

    info_handler = TestHandler()
    info_handler.setLevel("INFO")
    debug_handler = TestHandler()
    debug_handler.setLevel("DEBUG")

    log = logging.getLogger("bluesky")
    log.setLevel("DEBUG")
    assert log.getEffectiveLevel() == 10
    sub_log = logging.getLogger("bluesky.test")
    log.addHandler(info_handler)
    log.addHandler(debug_handler)
    log.debug("test debug")
    log.info("test debug")
    assert len(info_handler.records) == 1
    assert len(debug_handler.records) == 2
    assert set(info_handler.records).issubset(debug_handler.records)

    info_handler.records.clear()
    debug_handler.records.clear()

    sub_log = logging.getLogger("bluesky.test")
    assert sub_log.getEffectiveLevel() == 10
    sub_log.debug("test debug")
    sub_log.info("test debug")
    assert len(info_handler.records) == 1
    assert len(debug_handler.records) == 2
    assert set(info_handler.records).issubset(debug_handler.records)



import functools
from collections import defaultdict
from time import time

import pytest
from event_model.documents.event import PartialEvent
from ophyd import Component as Cpt
from ophyd import Device
from ophyd.sim import NullStatus, StatusBase, TrivialFlyer

from bluesky import Msg
from bluesky.plan_stubs import (
    close_run,
    collect_while_completing,
    complete,
    complete_all,
    declare_stream,
    kickoff,
    kickoff_all,
    open_run,
    wait,
)
from bluesky.plans import count, fly
from bluesky.protocols import Preparable
from bluesky.run_engine import IllegalMessageSequence
from bluesky.tests import requires_ophyd
from bluesky.tests.utils import DocCollector


def collect_while_completing_plan(flyers, dets, stream_name: str = "test_stream", pre_declare: bool = True):
    yield from open_run()
    if pre_declare:
        yield from declare_stream(*dets, name=stream_name, collect=True)
    yield from collect_while_completing(flyers, dets, flush_period=0.1, stream_name=stream_name)
    yield from close_run()


def test_collect_while_completing_plan_trivial_case(RE):
    completed = []
    collected = []

    class StatusDoneAfterTenthCall(StatusBase):
        times_called = 0

        @property
        def done(self):
            self.times_called += 1
            return self.times_called >= 10

    class SlowFlyer:
        name = "trivial-flyer"
        custom_status = StatusDoneAfterTenthCall()

        def kickoff(self):
            return NullStatus()

        def complete(self):
            completed.append(self)
            return self.custom_status

    class TrivialDetector:
        name = "trivial-detector"
        times_collected = 0

        def describe_collect(self):
            collected.append(self)
            return {
                "times_collected": {
                    "dims": [],
                    "dtype": "number",
                    "shape": [],
                    "source": "times_collected",
                }
            }

        def collect(self):
            self.times_collected += 1
            yield PartialEvent(
                data={"times_collected": self.times_collected},
                timestamps={"times_collected": time()},
            )

    det = TrivialDetector()
    flyer = SlowFlyer()
    docs = defaultdict(list)

    def assert_emitted(docs: dict[str, list], **numbers: int):
        assert list(docs) == list(numbers)
        assert {name: len(d) for name, d in docs.items()} == numbers

    RE(collect_while_completing_plan([flyer], [det]), lambda name, doc: docs[name].append(doc))
    for idx, event in enumerate(docs["event_page"]):
        print(idx, event)
        assert "times_collected" in event["data"] and event["data"]["times_collected"] == [idx + 1]
    # The detector will be collected nine times as the flyer's done property is
    # checked once during the initial complete call, then the detector collects
    # every loop and checks if the flyer is done until it is checked for a tenth time
    assert_emitted(docs, start=1, descriptor=1, event_page=9, stop=1)
    # key should be removed from set when collection done
    assert not RE._seen_wait_and_move_on_keys
    assert det in collected
    assert flyer in completed


@requires_ophyd
def test_flyer_with_collect_asset_documents(RE):
    from ophyd.sim import det, new_trivial_flyer, trivial_flyer

    from bluesky.preprocessors import fly_during_wrapper

    assert hasattr(new_trivial_flyer, "collect_asset_docs")
    assert not hasattr(trivial_flyer, "collect_asset_docs")
    RE(fly_during_wrapper(count([det], num=5), [new_trivial_flyer, trivial_flyer]))


@requires_ophyd
def test_collect_uncollected_and_log_any_errors(RE):
    # test that if stopping one motor raises an error, we can carry on
    collected = {}

    from ophyd.sim import TrivialFlyer

    class DummyFlyerWithFlag(TrivialFlyer):
        def collect(self):
            collected[self.name] = True
            super().collect()

    class BrokenDummyFlyerWithFlag(DummyFlyerWithFlag):
        def collect(self):
            super().collect()
            raise Exception

    flyer1 = DummyFlyerWithFlag()
    flyer1.name = "flyer1"
    flyer2 = BrokenDummyFlyerWithFlag()
    flyer2.name = "flyer2"

    collected.clear()
    RE([Msg("open_run"), Msg("kickoff", flyer1), Msg("kickoff", flyer2)])
    assert "flyer1" in collected
    assert "flyer2" in collected

    collected.clear()
    RE([Msg("open_run"), Msg("kickoff", flyer2), Msg("kickoff", flyer1)])
    assert "flyer1" in collected
    assert "flyer2" in collected


@requires_ophyd
def test_flying_outside_a_run_is_illegal(RE, hw):
    flyer = hw.trivial_flyer

    # This is normal, legal usage.
    RE(
        [
            Msg("open_run"),
            Msg("kickoff", flyer, group="foo"),
            Msg("wait", group="foo"),
            Msg("complete", flyer, group="bar"),
            Msg("wait", group="bar"),
            Msg("collect", flyer),
            Msg("close_run"),
        ]
    )

    # This is normal, legal usage (partial collection).
    RE(
        [
            Msg("open_run"),
            Msg("kickoff", flyer, group="foo"),
            Msg("wait", group="foo"),
            Msg("collect", flyer),
            Msg("collect", flyer),
            Msg("collect", flyer),
            Msg("complete", flyer, group="bar"),
            Msg("wait", group="bar"),
            Msg("collect", flyer),
            Msg("collect", flyer),
            Msg("collect", flyer),
            Msg("close_run"),
        ]
    )

    # It is not legal to kickoff outside of a run.
    with pytest.raises(IllegalMessageSequence):
        RE([Msg("kickoff", flyer)])


@requires_ophyd
def test_flyer_descriptor(RE, hw):
    class Flyer(TrivialFlyer):
        def __init__(self, name):
            self.name = name
            self.detector = FlyerDetector(name="flyer-detector")

        def read_configuration(self):
            return self.detector.read_configuration()

        def describe_configuration(self):
            return self.detector.describe_configuration()

        def describe_collect(self):
            return {
                "primary": {
                    "data_key_1": {
                        "dims": [],
                        "dtype": "string",
                        "shape": [],
                        "source": "",
                    },
                    "data_key_2": {
                        "dims": [],
                        "dtype": "number",
                        "shape": [],
                        "source": "",
                    },
                },
                "secondary": {
                    "data_key_3": {
                        "dims": [],
                        "dtype": "string",
                        "shape": [],
                        "source": "",
                    },
                    "data_key_4": {
                        "dims": [],
                        "dtype": "number",
                        "shape": [],
                        "source": "",
                    },
                },
            }

        def collect(self):
            yield {
                "data": {"data_key_1": "1", "data_key_2": 2},
                "timestamps": {"data_key_1": 0, "data_key_2": 0},
                "time": 0,
            }

            yield {
                "data": {"data_key_3": "3", "data_key_4": 4},
                "timestamps": {"data_key_3": 0, "data_key_4": 0},
                "time": 0,
            }

    flyers = [Flyer(name="flyer"), TrivialFlyer()]
    descriptors = dict()  # noqa: C408

    RE(
        fly(flyers),
        {"descriptor": lambda name, doc: descriptors.update({doc["name"]: doc})},
    )

    primary_descriptor = descriptors["primary"]
    assert primary_descriptor["configuration"]["flyer"] == {
        "data": {
            "config_key_1": "1",
            "config_key_2": 2,
            "config_key_3": "3",
            "config_key_4": 4,
        },
        "timestamps": {
            "config_key_1": 1,
            "config_key_2": 2,
            "config_key_3": 3,
            "config_key_4": 4,
        },
        "data_keys": {
            "config_key_1": {"dtype": "string", "shape": [], "source": "PV:Config:1"},
            "config_key_2": {"dtype": "number", "shape": [], "source": "PV:Config:2"},
            "config_key_3": {"dtype": "string", "shape": [], "source": "PV:Config:3"},
            "config_key_4": {"dtype": "number", "shape": [], "source": "PV:Config:4"},
        },
    }

    assert "flyer" in primary_descriptor["object_keys"]

    secondary_descriptor = descriptors["secondary"]
    assert len(secondary_descriptor["configuration"]["flyer"]["data"]) == 4
    assert secondary_descriptor["configuration"] == primary_descriptor["configuration"]

    assert "flyer" in secondary_descriptor["object_keys"]

    trivial_flyer_descriptor = descriptors["stream_name"]
    print(f"trivial flyer descriptor: {trivial_flyer_descriptor}")
    assert len(trivial_flyer_descriptor["configuration"]) == 1
    assert "trivial_flyer" in trivial_flyer_descriptor["object_keys"]


@requires_ophyd
def test_device_flyer_descriptor(RE, hw):
    # TrivialFlyer is not a Device
    flyers = [FlyerDevice(name="flyer-detector"), TrivialFlyer()]
    descriptors = dict()  # noqa: C408

    RE(
        fly(flyers),
        {"descriptor": lambda name, doc: descriptors.update({doc["name"]: doc})},
    )

    primary_descriptor = descriptors["primary"]
    print(f"primary descriptor: {primary_descriptor}")
    assert len(primary_descriptor["configuration"]) == 1
    assert primary_descriptor["configuration"]["flyer-detector"] == {
        "data": {
            "config_key_1": "1",
            "config_key_2": 2,
            "config_key_3": "3",
            "config_key_4": 4,
        },
        "timestamps": {
            "config_key_1": 1,
            "config_key_2": 2,
            "config_key_3": 3,
            "config_key_4": 4,
        },
        "data_keys": {
            "config_key_1": {"dtype": "string", "shape": [], "source": "PV:Config:1"},
            "config_key_2": {"dtype": "number", "shape": [], "source": "PV:Config:2"},
            "config_key_3": {"dtype": "string", "shape": [], "source": "PV:Config:3"},
            "config_key_4": {"dtype": "number", "shape": [], "source": "PV:Config:4"},
        },
    }

    secondary_descriptor = descriptors["secondary"]
    print(f"secondary_descriptor: {secondary_descriptor}")
    assert len(secondary_descriptor["configuration"]["flyer-detector"]["data"]) == 4
    assert secondary_descriptor["configuration"] == primary_descriptor["configuration"]

    trivial_flyer_descriptor = descriptors["stream_name"]
    print(f"trivial flyer descriptor: {trivial_flyer_descriptor}")
    assert len(trivial_flyer_descriptor["configuration"]) == 1
    assert "trivial_flyer" in trivial_flyer_descriptor["object_keys"]


def test_device_redundent_config_reading(RE):
    flyer = FlyerDevice(name="flyer-detector")
    RE(
        [
            Msg("open_run"),
            Msg("kickoff", flyer, group="foo"),
            Msg("wait", group="foo"),
            Msg("collect", flyer),
            Msg("collect", flyer),
            Msg("collect", flyer),
            Msg("complete", flyer, group="bar"),
            Msg("wait", group="bar"),
            Msg("collect", flyer),
            Msg("close_run"),
        ]
    )
    assert flyer.call_counts["collect"] == 4
    assert flyer.call_counts["describe_collect"] == 1
    assert flyer.call_counts["read_configuration"] == 1
    assert flyer.call_counts["describe_configuration"] == 1
    assert flyer.call_counts["kickoff"] == 1
    assert flyer.call_counts["complete"] == 1


class FlyerDetector(Device):
    def __init__(self, name, *args, **kwargs):
        super().__init__(name=name, *args, **kwargs)  # noqa: B026

    def describe_configuration(self):
        return {
            "config_key_1": {"dtype": "string", "shape": [], "source": "PV:Config:1"},
            "config_key_2": {"dtype": "number", "shape": [], "source": "PV:Config:2"},
            "config_key_3": {"dtype": "string", "shape": [], "source": "PV:Config:3"},
            "config_key_4": {"dtype": "number", "shape": [], "source": "PV:Config:4"},
        }

    def read_configuration(self):
        return {
            "config_key_1": {"value": "1", "timestamp": 1},
            "config_key_2": {"value": 2, "timestamp": 2},
            "config_key_3": {"value": "3", "timestamp": 3},
            "config_key_4": {"value": 4, "timestamp": 4},
        }


def call_counter(func):
    @functools.wraps(func)
    def inner(self, *args, **kwargs):
        self.call_counts[func.__name__] += 1
        return func(self, *args, **kwargs)

    return inner


class FlyerDevice(Device):
    detector = Cpt(FlyerDetector, name="flyer-detector")

    def __init__(self, *args, **kwargs):
        self.call_counts = defaultdict(int)
        super().__init__(*args, **kwargs)

    @call_counter
    def kickoff(self):
        return NullStatus()

    @call_counter
    def complete(self):
        return NullStatus()

    def stop(self, *, success=False):
        pass

    @call_counter
    def describe_collect(self):
        return {
            "primary": {
                "data_key_1": {
                    "dims": [],
                    "dtype": "string",
                    "shape": [],
                    "source": "",
                },
                "data_key_2": {
                    "dims": [],
                    "dtype": "number",
                    "shape": [],
                    "source": "",
                },
            },
            "secondary": {
                "data_key_3": {
                    "dims": [],
                    "dtype": "string",
                    "shape": [],
                    "source": "",
                },
                "data_key_4": {
                    "dims": [],
                    "dtype": "number",
                    "shape": [],
                    "source": "",
                },
            },
        }

    @call_counter
    def collect(self):
        yield {
            "data": {"data_key_1": "", "data_key_2": 0},
            "timestamps": {"data_key_1": 0, "data_key_2": 0},
            "time": 0,
        }

        yield {
            "data": {"data_key_3": "", "data_key_4": 0},
            "timestamps": {"data_key_3": 0, "data_key_4": 0},
            "time": 0,
        }

    @call_counter
    def read_configuration(self):
        return super().read_configuration()

    @call_counter
    def describe_configuration(self):
        return super().describe_configuration()


def test_describe_config_optional(RE):
    class Simple:
        """A trivial flyer that omits the configuration methods"""

        name = "simple_flyer"
        parent = None

        def kickoff(self):
            return NullStatus()

        def describe_collect(self):
            return {"stream_name": {}}

        def complete(self):
            return NullStatus()

        def collect(self):
            for i in range(100):
                yield {"data": {}, "timestamps": {}, "time": i, "seq_num": i}

        def stop(self, *, success=False):
            pass

    d = DocCollector()

    flyer = Simple()

    RE(
        [
            Msg("open_run"),
            Msg("kickoff", flyer, group="foo"),
            Msg("wait", group="foo"),
            Msg("complete", flyer, group="bar"),
            Msg("wait", group="bar"),
            Msg("collect", flyer),
            Msg("close_run"),
        ],
        d.insert,
    )

    ((desc,),) = d.descriptor.values()
    assert desc["name"] == "stream_name"

    assert "simple_flyer" in desc["object_keys"]
    assert "simple_flyer" in desc["configuration"]


def test_prepare(RE):
    """Tests that obj.prepare is correctly called for a fly scan."""

    class PreparableFlyer(FlyerDevice, Preparable):
        @call_counter
        def prepare(self, value):
            self.value = value
            return NullStatus()

    flyer = PreparableFlyer(name="flyer")

    # We can prepare the flyer even if no run is open,
    RE([Msg("prepare", flyer, "abcdefg")])

    assert flyer.value == "abcdefg"

    fly_scan_sequence = [
        Msg("open_run"),
        Msg("prepare", flyer, 123),
        Msg("kickoff", flyer, group="foo"),
        Msg("wait", group="foo"),
        Msg("complete", flyer, group="bar"),
        Msg("wait", group="bar"),
        Msg("collect", flyer),
        Msg("close_run"),
    ]

    RE(fly_scan_sequence)

    assert flyer.value == 123

    assert flyer.call_counts["prepare"] == 2  # called once at start of test
    assert flyer.call_counts["kickoff"] == 1
    assert flyer.call_counts["complete"] == 1
    assert flyer.call_counts["collect"] == 1


def test_kickoff_complete_all_single_device(RE):
    flyer1 = FlyerDevice(name="flyer1")

    def fly_plan():
        yield from open_run()
        yield from kickoff_all(flyer1, group="foo")
        yield from wait(group="foo")
        yield from complete_all(flyer1, group="bar")
        yield from wait(group="bar")
        yield from close_run()

    RE(fly_plan())

    assert flyer1.call_counts["kickoff"] == 1
    assert flyer1.call_counts["complete"] == 1


def test_kickoff_complete_all_take_multiple_devices(RE):
    flyer1 = FlyerDevice(name="flyer1")
    flyer2 = FlyerDevice(name="flyer2")

    def fly_plan():
        yield from open_run()
        yield from kickoff_all(flyer1, flyer2, group="foo")
        yield from wait(group="foo")
        yield from complete_all(flyer1, flyer2, group="bar")
        yield from wait(group="bar")
        yield from close_run()

    RE(fly_plan())

    assert flyer1.call_counts["kickoff"] == 1
    assert flyer2.call_counts["kickoff"] == 1
    assert flyer1.call_counts["complete"] == 1
    assert flyer2.call_counts["complete"] == 1


def test_kickoff_all_complete_seperately(RE):
    flyer1 = FlyerDevice(name="flyer1")
    flyer2 = FlyerDevice(name="flyer2")

    def fly_plan():
        yield from open_run()
        yield from kickoff_all(flyer1, flyer2, group="foo")
        yield from wait(group="foo")
        yield from complete(flyer1, group="bar")
        yield from complete(flyer2, group="bar")
        yield from wait(group="bar")
        yield from close_run()

    RE(fly_plan())

    assert flyer1.call_counts["kickoff"] == 1
    assert flyer2.call_counts["kickoff"] == 1
    assert flyer1.call_counts["complete"] == 1
    assert flyer2.call_counts["complete"] == 1


def test_kickoff_seperately_complete_all(RE):
    flyer1 = FlyerDevice(name="flyer1")
    flyer2 = FlyerDevice(name="flyer2")

    def fly_plan():
        yield from open_run()
        yield from kickoff(flyer1, group="foo")
        yield from kickoff(flyer2, group="foo")
        yield from wait(group="foo")
        yield from complete_all(flyer1, flyer2, group="bar")
        yield from wait(group="bar")
        yield from close_run()

    RE(fly_plan())

    assert flyer1.call_counts["kickoff"] == 1
    assert flyer2.call_counts["kickoff"] == 1
    assert flyer1.call_counts["complete"] == 1
    assert flyer2.call_counts["complete"] == 1



from enum import Enum

import pytest

from bluesky._vendor.super_state_machine import machines


class StatesEnum(Enum):
    ONE = "one"
    TWO = "two"
    THREE = "three"


class OtherEnum(Enum):
    ONE = "one"


def test_state_machine_is_always_scalar():
    class Machine(machines.StateMachine):
        class States(Enum):
            ONE = "one"
            TWO = "two"
            THREE = "three"

        state = States.ONE

    sm = Machine()
    assert sm.state == "one"


def test_state_machine_allow_scalars_on_init():
    class Machine(machines.StateMachine):
        class States(Enum):
            ONE = "one"
            TWO = "two"
            THREE = "three"

        state = "two"

    sm = Machine()
    assert sm.is_two is True


def test_state_machine_accepts_enums_only_from_proper_source():
    class OtherEnum(Enum):
        FOUR = "four"

    with pytest.raises(ValueError):

        class Machine(machines.StateMachine):
            class States(Enum):
                ONE = "one"
                TWO = "two"
                THREE = "three"

            state = OtherEnum.FOUR


def test_state_machine_doesnt_allow_wrong_scalars():
    with pytest.raises(ValueError):

        class Machine(machines.StateMachine):
            class States(Enum):
                ONE = "one"
                TWO = "two"
                THREE = "three"

            state = "four"


def test_state_machine_accepts_only_unique_enums():
    with pytest.raises(ValueError):

        class Machine(machines.StateMachine):
            class States(Enum):
                ONE = "one"
                TWO = "one"
                THREE = "three"


def test_state_machine_allows_to_change_and_check_state():
    class Machine(machines.StateMachine):
        state = "one"

        class States(Enum):
            ONE = "one"
            TWO = "two"
            THREE = "three"

    sm = Machine()
    assert sm.is_("one") is True
    assert sm.state == "one"
    assert sm.is_("two") is False
    sm.set_("two")
    assert sm.is_("two") is True


def test_state_machine_allows_to_change_and_check_state_by_methods():
    class Machine(machines.StateMachine):
        state = "one"

        class States(Enum):
            ONE = "one"
            TWO = "two"
            THREE = "three"

    sm = Machine()
    assert sm.is_one is True
    assert sm.state == "one"
    assert sm.is_two is False
    sm.set_two()
    assert sm.is_two is True


def test_name_collistion_for_checker_raises_exception():
    with pytest.raises(ValueError):

        class Machine(machines.StateMachine):
            class States(Enum):
                ONE = "one"
                TWO = "two"
                THREE = "three"

            def is_one():
                pass


def test_name_collistion_for_setter_raises_exception():
    with pytest.raises(ValueError):

        class Machine(machines.StateMachine):
            class States(Enum):
                ONE = "one"
                TWO = "two"
                THREE = "three"

            def set_one():
                pass


def test_states_enum_can_be_predefined():
    class Machine(machines.StateMachine):
        States = StatesEnum
        state = StatesEnum.ONE

    sm = Machine()
    assert sm.is_one is True


def test_disallow_empty():
    class Machine(machines.StateMachine):
        States = StatesEnum
        state = StatesEnum.ONE

    sm = Machine()
    with pytest.raises(AttributeError):
        del sm.state


def test_states_enum_is_always_given():
    with pytest.raises(ValueError):

        class Machine(machines.StateMachine):
            pass


def test_states_enum_is_always_enum():
    with pytest.raises(ValueError):

        class Machine(machines.StateMachine):
            States = "something"


def test_disallow_empty_without_initial_value():
    with pytest.raises(ValueError):

        class Machine(machines.StateMachine):
            States = StatesEnum

            class Meta:
                allow_empty = False


def test_initial_value():
    class Machine(machines.StateMachine):
        States = StatesEnum
        state = StatesEnum.ONE

        class Meta:
            initial_state = StatesEnum.TWO

    sm = Machine()
    assert sm.state == "two"
    assert sm.is_two is True


def test_wrong_initial_value_from_class_is_ignored():
    class Machine(machines.StateMachine):
        States = StatesEnum
        state = "wrong"

        class Meta:
            initial_state = StatesEnum.TWO

    sm = Machine()
    assert sm.state == "two"
    assert sm.is_two is True


def test_initial_value_from_meta_and_disallowed_empty():
    class Machine(machines.StateMachine):
        States = StatesEnum

        class Meta:
            allow_empty = False
            initial_state = StatesEnum.TWO

    sm = Machine()
    assert sm.state == "two"
    assert sm.is_two is True


def test_disallow_assignation_of_wrong_value():
    class Machine(machines.StateMachine):
        state = "one"

        class States(Enum):
            ONE = "one"
            TWO = "two"
            THREE = "three"

    sm = Machine()
    sm.set_one()
    assert sm.is_one is True
    sm.set_(Machine.States.TWO)
    assert sm.is_two is True
    with pytest.raises(ValueError):
        sm.set_(StatesEnum.TWO)
    with pytest.raises(ValueError):
        sm.set_("four")


def test_checker_getter_and_setter_wrong_values_and_enums():
    class Machine(machines.StateMachine):
        States = StatesEnum
        state = "one"

    sm = Machine()

    sm.is_("one")
    sm.can_be_("one")
    sm.set_("one")

    with pytest.raises(ValueError):
        sm.is_("five")
    with pytest.raises(ValueError):
        sm.is_(OtherEnum.ONE)
    with pytest.raises(ValueError):
        sm.can_be_("five")
    with pytest.raises(ValueError):
        sm.can_be_(OtherEnum.ONE)
    with pytest.raises(ValueError):
        sm.set_("five")
    with pytest.raises(ValueError):
        sm.set_(OtherEnum.ONE)


def test_get_actual_state_as_enum():
    class Machine(machines.StateMachine):
        States = StatesEnum
        state = "one"

    sm = Machine()
    assert sm.actual_state is StatesEnum.ONE
    sm.set_two()
    assert sm.actual_state is StatesEnum.TWO


def test_actual_state_name_collision():
    with pytest.raises(ValueError):

        class Machine(machines.StateMachine):
            States = StatesEnum

            def actual_state():
                pass


def test_actual_state_name_collision_with_generated_methods():
    with pytest.raises(ValueError):

        class Machine(machines.StateMachine):
            States = StatesEnum

            class Meta:
                named_transitions = [
                    ("actual_state", "one"),
                ]


def test_as_enum():
    class Machine(machines.StateMachine):
        States = StatesEnum
        state = "one"

    sm = Machine()
    assert sm.as_enum is StatesEnum.ONE
    sm.set_two()
    assert sm.as_enum is StatesEnum.TWO


def test_as_enum_name_collision():
    with pytest.raises(ValueError):

        class Machine(machines.StateMachine):
            States = StatesEnum

            def as_enum():
                pass


def test_as_enum_name_collision_with_generated_methods():
    with pytest.raises(ValueError):

        class Machine(machines.StateMachine):
            States = StatesEnum

            class Meta:
                named_transitions = [
                    ("as_enum", "one"),
                ]


def test_custom_states_enum():
    class Machine(machines.StateMachine):
        trololo = StatesEnum
        state = "one"

        class Meta:
            states_enum_name = "trololo"

    machine = Machine()
    assert machine.is_one is True


def test_initial_state_is_required():
    with pytest.raises(ValueError):

        class Machine(machines.StateMachine):
            States = StatesEnum

            class Meta:
                pass


def test_state_machine_allows_only_strings():
    with pytest.raises(ValueError):

        class Machine(machines.StateMachine):
            class States(Enum):
                ONE = 1
                TWO = "two"
                THREE = "three"



import time

import numpy as np
import pytest

from bluesky import Msg
from bluesky.plan_stubs import declare_stream, trigger_and_read
from bluesky.plans import ramp_plan
from bluesky.tests import requires_ophyd
from bluesky.tests.utils import DocCollector
from bluesky.utils import RampFail


@requires_ophyd
def test_ramp(RE):
    from ophyd import StatusBase
    from ophyd.positioner import SoftPositioner
    from ophyd.sim import SynGauss

    import bluesky.utils as bsu

    RE.msg_hook = bsu.ts_msg_hook

    tt = SoftPositioner(name="mot")
    tt.set(0)
    dd = SynGauss("det", tt, "mot", 0, 3)

    st = StatusBase()

    def kickoff():
        yield Msg("null")
        for j, v in enumerate(np.linspace(-5, 5, 10)):
            RE.loop.call_later(0.1 * j, lambda v=v: tt.set(v))
        RE.loop.call_later(1.2, st._finished)
        print("YOLO")
        return st

    first = True

    def inner_plan():
        nonlocal first
        if first:
            yield from declare_stream(dd, name="primary")
            first = False

        yield from trigger_and_read([dd])

    g = ramp_plan(kickoff(), tt, inner_plan, period=0.08)
    db = DocCollector()
    RE.subscribe(db.insert)
    rs = RE(g)
    if RE.call_returns_result:
        uid = rs.run_start_uids[0]
    else:
        uid = rs[0]
    assert db.start[0]["uid"] == uid
    assert len(db.descriptor[uid]) == 2
    descs = {d["name"]: d for d in db.descriptor[uid]}

    assert set(descs) == set(["primary", "mot_monitor"])  # noqa: C405

    primary_events = db.event[descs["primary"]["uid"]]
    assert len(primary_events) > 11

    monitor_events = db.event[descs["mot_monitor"]["uid"]]
    # the 10 from the updates, 1 from 'run at subscription time'
    assert len(monitor_events) == 11


@requires_ophyd
def test_timeout(RE):
    from ophyd import StatusBase
    from ophyd.positioner import SoftPositioner
    from ophyd.sim import SynGauss

    mot = SoftPositioner(name="mot")
    mot.set(0)
    det = SynGauss("det", mot, "mot", 0, 3)

    def kickoff():
        # This process should take a total of 5 seconds, but it will be
        # interrupted after 0.1 seconds via the timeout keyword passed to
        # ramp_plan below.
        yield Msg("null")
        for j in range(100):
            RE.loop.call_later(0.05 * j, lambda j=j: mot.set(j))

        return StatusBase()

    first = True

    def inner_plan():
        nonlocal first
        if first:
            yield from declare_stream(det, name="primary")
            first = False

        yield from trigger_and_read([det])

    g = ramp_plan(kickoff(), mot, inner_plan, period=0.01, timeout=0.1)

    start = time.time()
    with pytest.raises(RampFail):
        RE(g)
    stop = time.time()
    elapsed = stop - start

    # This test has a tendency to randomly fail, so we're giving it plenty of
    # time to run. 4, though much greater than 0.1, is still less than 5.
    assert 0.1 < elapsed < 4



import asyncio
import threading
import time as ttime
from collections import defaultdict
from types import SimpleNamespace

import pytest

import bluesky.plans as bp
from bluesky import Msg, RunEngineInterrupted
from bluesky.plan_stubs import (
    abs_set,
    caching_repeater,
    checkpoint,
    clear_checkpoint,
    close_run,
    collect,
    configure,
    create,
    deferred_pause,
    drop,
    install_suspender,
    kickoff,
    locate,
    monitor,
    mv,
    mvr,
    null,
    one_1d_step,
    one_nd_step,
    one_shot,
    open_run,
    pause,
    rd,
    read,
    rel_set,
    remove_suspender,
    repeat,
    repeater,
    save,
    sleep,
    stage,
    stage_all,
    stop,
    subscribe,
    trigger,
    trigger_and_read,
    unmonitor,
    unstage,
    unstage_all,
    unsubscribe,
    wait,
    wait_for,
)
from bluesky.plans import count, inner_product_scan, rel_scan, scan
from bluesky.preprocessors import (
    configure_count_time_wrapper,
    finalize_decorator,
    finalize_wrapper,
    fly_during_decorator,
    fly_during_wrapper,
    inject_md_wrapper,
    lazily_stage_wrapper,
    monitor_during_decorator,
    monitor_during_wrapper,
    relative_set_wrapper,
    reset_positions_wrapper,
    subs_decorator,
    subs_wrapper,
    suspend_wrapper,
)
from bluesky.protocols import Descriptor, Locatable, Location, Readable, Reading, Status
from bluesky.utils import IllegalMessageSequence, all_safe_rewind


@pytest.mark.parametrize(
    "plan,plan_args,plan_kwargs,msgs",
    [
        (create, (), {}, [Msg("create", name="primary")]),
        (create, ("custom_name",), {}, [Msg("create", name="custom_name")]),
        (save, (), {}, [Msg("save")]),
        (drop, (), {}, [Msg("drop")]),
        (read, ("det",), {}, [Msg("read", "det")]),
        (locate, (), {}, [Msg("locate", squeeze=True)]),
        (
            locate,
            ("foo",),
            {"squeeze": False},
            [Msg("locate", "foo", squeeze=False)],
        ),
        (monitor, ("foo",), {}, [Msg("monitor", "foo", name=None)]),
        (monitor, ("foo",), {"name": "c"}, [Msg("monitor", "foo", name="c")]),
        (unmonitor, ("foo",), {}, [Msg("unmonitor", "foo")]),
        (null, (), {}, [Msg("null")]),
        (stop, ("foo",), {}, [Msg("stop", "foo")]),
        (abs_set, ("det", 5), {}, [Msg("set", "det", 5, group=None)]),
        (abs_set, ("det", 5), {"group": "A"}, [Msg("set", "det", 5, group="A")]),
        (
            abs_set,
            ("det", 5),
            {"group": "A", "wait": True},
            [Msg("set", "det", 5, group="A"), Msg("wait", None, group="A")],
        ),
        (rel_set, ("det", 5), {}, [Msg("read", "det"), Msg("set", "det", 5, group=None)]),
        (rel_set, ("det", 5), {"group": "A"}, [Msg("read", "det"), Msg("set", "det", 5, group="A")]),
        (
            rel_set,
            ("det", 5),
            {"group": "A", "wait": True},
            [Msg("read", "det"), Msg("set", "det", 5, group="A"), Msg("wait", None, group="A")],
        ),
        (
            mv,
            ("motor1", 1, "motor2", 2),
            {"group": "A"},
            [
                Msg("set", "motor1", 1, group="A"),
                Msg("set", "motor2", 2, group="A"),
                Msg("wait", None, group="A", timeout=None),
            ],
        ),
        (
            mv,
            ("motor1", 1, "motor2", 2),
            {"group": "A", "timeout": 42},
            [
                Msg("set", "motor1", 1, group="A"),
                Msg("set", "motor2", 2, group="A"),
                Msg("wait", None, group="A", timeout=42),
            ],
        ),
        (trigger, ("det",), {}, [Msg("trigger", "det", group=None)]),
        (trigger, ("det",), {"group": "A"}, [Msg("trigger", "det", group="A")]),
        (sleep, (2,), {}, [Msg("sleep", None, 2)]),
        (wait, (), {}, [Msg("wait", None, error_on_timeout=True, group=None, timeout=None, watch=())]),
        (wait, ("A",), {}, [Msg("wait", None, group="A", error_on_timeout=True, timeout=None, watch=())]),
        (checkpoint, (), {}, [Msg("checkpoint")]),
        (clear_checkpoint, (), {}, [Msg("clear_checkpoint")]),
        (pause, (), {}, [Msg("pause", None, defer=False)]),
        (deferred_pause, (), {}, [Msg("pause", None, defer=True)]),
        (kickoff, ("foo",), {}, [Msg("kickoff", "foo", group=None)]),
        (kickoff, ("foo",), {"custom": 5}, [Msg("kickoff", "foo", group=None, custom=5)]),
        (collect, ("foo",), {}, [Msg("collect", "foo", stream=False, return_payload=True, name=None)]),
        (configure, ("det", 1), {"a": 2}, [Msg("configure", "det", 1, a=2)]),
        (stage, ("det",), {}, [Msg("stage", "det", group=None)]),
        (stage, ("det",), {"group": "A"}, [Msg("stage", "det", group="A")]),
        (unstage, ("det",), {}, [Msg("unstage", "det", group=None)]),
        (unstage, ("det",), {"group": "A"}, [Msg("unstage", "det", group="A")]),
        (subscribe, ("all", "func_placeholder"), {}, [Msg("subscribe", None, "func_placeholder", "all")]),
        (unsubscribe, (1,), {}, [Msg("unsubscribe", None, token=1)]),
        (install_suspender, (1,), {}, [Msg("install_suspender", None, 1)]),
        (remove_suspender, (1,), {}, [Msg("remove_suspender", None, 1)]),
        (open_run, (), {}, [Msg("open_run")]),
        (open_run, (), {"md": {"a": 1}}, [Msg("open_run", a=1)]),
        (close_run, (), {}, [Msg("close_run", reason=None, exit_status=None)]),
        (wait_for, (["fut1", "fut2"],), {}, [Msg("wait_for", None, ["fut1", "fut2"])]),
    ],
)
def test_stub_plans(plan, plan_args, plan_kwargs, msgs, hw):
    # de-reference
    plan_args = tuple(getattr(hw, v, v) if isinstance(v, str) else v for v in plan_args)
    plan_kwargs = {k: getattr(hw, v, v) if isinstance(v, str) else v for k, v in plan_kwargs.items()}
    msgs = [
        Msg(m.command, getattr(hw, m.obj, m.obj) if isinstance(m.obj, str) else m.obj, *m.args, **m.kwargs)
        for m in msgs
    ]
    assert list(plan(*plan_args, **plan_kwargs)) == msgs


@pytest.mark.parametrize(("timeout", "should_fail"), [(0, True), (1, False), (None, False)])
def test_mv_timeout(RE, hw, timeout, should_fail):
    sig = hw.motor
    sig.delay = 0.01

    def tester(obj):
        try:
            yield from mv(obj, 1, timeout=timeout)
        except TimeoutError:
            assert should_fail
        else:
            assert not should_fail

        # This needs to happen so all Futures can be cleared inside RE.
        yield from sleep(obj.delay)

    RE(tester(sig))


def test_mvr(RE, hw):
    # special-case mvr because the value cannot be pre-defined in test_stub_plans
    # move motors first to ensure that movement is relative, not absolute
    hw.motor1.set(10)
    hw.motor2.set(10)
    actual = []
    RE.msg_hook = lambda msg: actual.append(msg)
    RE(mvr(hw.motor1, 1, hw.motor2, 2))
    actual = list(mv(hw.motor1, 1, hw.motor2, 2))
    strip_group(actual)
    for msg in actual[:2]:
        msg.command == "set"  # noqa: B015
    assert set([msg.obj for msg in actual[:2]]) == set([hw.motor1, hw.motor2])  # noqa: C403, C405
    assert actual[2] == Msg("wait", None, timeout=None)


def test_locatable_message_multiple_objects(RE, hw):
    class AsyncLocatable(Locatable):
        value = 1.0

        def set(self, value) -> Status:
            self.position = value + 0.1
            # We don't actually need a motor, but use one just to give us a Status
            return hw.motor.set(value)

        async def locate(self) -> Location:
            # Grab the original value
            value = AsyncLocatable.value
            # Let any other coros get a look in
            await asyncio.sleep(0)
            # Increment the next value and return the original value
            AsyncLocatable.value += 1
            return dict(setpoint=value, readback=value)  # noqa: C408

    rds = []
    one = AsyncLocatable()
    two = AsyncLocatable()

    def multi_rd():
        rds.append((yield Msg("locate", one, squeeze=False)))
        rds.append((yield Msg("locate", two)))
        rds.append((yield Msg("locate", one, two)))
        rds.append((yield Msg("locate", one)))

    RE(multi_rd())
    assert rds == [
        [dict(setpoint=1.0, readback=1.0)],  # noqa: C408
        dict(setpoint=2.0, readback=2.0),  # noqa: C408
        # Check they happened at the same time, so have the same value
        [dict(setpoint=3.0, readback=3.0), dict(setpoint=3.0, readback=3.0)],  # noqa: C408
        # And now it will skip one as they both incremented above
        dict(setpoint=5.0, readback=5.0),  # noqa: C408
    ]


def test_rd_locatable(RE, hw):
    class Jittery(Readable, Locatable):
        def describe(self) -> dict[str, Descriptor]:
            return dict(x=dict(source="dummy", dtype="number", shape=[]))  # noqa: C408

        def read(self) -> dict[str, Reading]:
            return dict(x=dict(value=1.2, timestamp=0.0))  # noqa: C408

        def locate(self) -> Location:
            return dict(setpoint=1.0, readback=1.1)  # noqa: C408

        def set(self, value) -> Status:
            self.position = value + 0.1
            # We don't actually need a motor, but use one just to give us a Status
            return hw.motor.set(value)

        name = ""

    jittery = Jittery()
    rds = []

    def store_rd():
        value = yield from rd(jittery)
        rds.append(value)

    RE(store_rd())
    assert rds == [1.1]


def test_mvr_with_location(RE, hw):
    class AlmostThereMotor:
        # This is the readback
        position = 10.1
        parent = None
        name = "Almost"

        def set(self, value) -> Status:
            self.position = value + 0.1
            # We don't actually need a motor, but use one just to give us a Status
            return hw.motor.set(value)

        async def locate(self) -> Location:
            await asyncio.sleep(0.1)
            return dict(setpoint=self.position - 0.1, readback=self.position)  # noqa: C408

    m = AlmostThereMotor()
    assert isinstance(m, Locatable)
    actual = []
    RE.msg_hook = lambda msg: actual.append(msg)
    RE(mvr(m, 1))
    RE(mvr(m, 2))
    assert [x.command for x in actual] == ["locate", "set", "wait"] * 2
    assert [x.obj for x in actual] == [m, m, None, m, m, None]
    assert actual[1].args == (11,)
    assert actual[4].args == (13,)


def test_mvr_with_timeout(hw):
    # special-case mv because the group is not configurable
    # move motors first to ensure that movement is absolute, not relative
    actual = list(mvr(hw.motor1, 1, hw.motor2, 2, timeout=42))
    for msg in actual[:2]:
        assert msg.command == "set"  # noqa: B015

    assert actual[2].kwargs["timeout"] == 42


def strip_group(plan):
    for msg in plan:
        msg.kwargs.pop("group", None)


def test_monitor_during_wrapper(hw):
    det = hw.det

    def plan():
        # can't use 2 * [Msg('open_run'), Msg('null'), Msg('close_run')]
        # because plan_mutator sees the same ids twice and skips them
        yield from [Msg("open_run"), Msg("null"), Msg("close_run"), Msg("open_run"), Msg("null"), Msg("close_run")]

    processed_plan = list(monitor_during_wrapper(plan(), [det]))
    expected = 2 * [
        Msg("open_run"),
        # inserted
        Msg("monitor", det, name=(det.name + "_monitor")),
        Msg("null"),
        Msg("unmonitor", det),  # inserted
        Msg("close_run"),
    ]

    strip_group(processed_plan)
    assert processed_plan == expected

    processed_plan = list(monitor_during_decorator([det])(plan)())
    strip_group(processed_plan)
    assert processed_plan == expected


def test_descriptor_layout_from_monitor(RE, hw):
    collector = []
    det = hw.rand
    det.start_simulation()

    def collect(name, doc):
        if name == "descriptor":
            collector.append(doc)

    RE([Msg("open_run"), Msg("monitor", det, name=det.name), Msg("unmonitor", det), Msg("close_run")], collect)

    (descriptor,) = collector
    assert descriptor["object_keys"] == {det.name: list(det.describe().keys())}
    assert descriptor["data_keys"] == {k: {**v, "object_name": det.name} for k, v in det.describe().items()}
    conf = descriptor["configuration"][det.name]
    assert conf["data_keys"] == det.describe_configuration()
    vals = {key: val["value"] for key, val in det.read_configuration().items()}
    timestamps = {key: val["timestamp"] for key, val in det.read_configuration().items()}
    assert conf["data"] == vals
    assert conf["timestamps"].keys() == timestamps.keys()
    for val in conf["timestamps"].values():
        assert type(val) is float  # can't check actual value; time has passed  # noqa: E721


def test_fly_during():
    def plan():
        # can't use 2 * [Msg('open_run'), Msg('null'), Msg('close_run')]
        # because plan_mutator sees the same ids twice and skips them
        yield from [Msg("open_run"), Msg("null"), Msg("close_run"), Msg("open_run"), Msg("null"), Msg("close_run")]

    processed_plan = list(fly_during_wrapper(plan(), ["foo"]))
    expected = 2 * [
        Msg("open_run"),
        Msg("kickoff", "foo"),
        Msg("wait"),  # inserted
        Msg("null"),
        Msg("complete", "foo"),
        Msg("wait"),  # inserted
        Msg("collect", "foo"),  # inserted
        Msg("close_run"),
    ]

    strip_group(processed_plan)
    assert processed_plan == expected

    processed_plan = list(fly_during_decorator(["foo"])(plan)())
    strip_group(processed_plan)
    assert processed_plan == expected


def test_lazily_stage(hw: SimpleNamespace):
    det1, det2 = hw.det1, hw.det2

    def plan():
        yield from [Msg("read", det1), Msg("read", det1), Msg("read", det2)]

    processed_plan = list(lazily_stage_wrapper(plan()))

    expected_plan: list[Msg] = [
        Msg("stage", det1),
        Msg("read", det1),
        Msg("read", det1),
        Msg("stage", det2),
        Msg("read", det2),
        Msg("unstage", det2),
        Msg("unstage", det1),
    ]

    # Prevent issue with unstage_all creating a randomly assigned group
    assert len(processed_plan) == len(expected_plan)
    for i in range(len(expected_plan)):
        expected, actual = expected_plan[i], processed_plan[i]
        assert actual.command == expected.command
        assert actual.obj == expected.obj
    groups = {msg.kwargs["group"] for msg in processed_plan if msg.command == "unstage"}
    assert len(groups) == 1  # All unstage messages are in the same group


def test_subs():
    def cb(name, doc):
        pass

    def plan(*args, **kwargs):
        # check that args to plan are passed through
        yield from [Msg("null", None, *args, **kwargs)]

    processed_plan = list(subs_wrapper(plan("test_arg", test_kwarg="val"), {"all": cb}))

    expected = [
        Msg("subscribe", None, cb, "all"),
        Msg("null", None, "test_arg", test_kwarg="val"),
        Msg("unsubscribe", token=None),
    ]

    assert processed_plan == expected

    processed_plan = list(subs_decorator({"all": cb})(plan)("test_arg", test_kwarg="val"))
    assert processed_plan == expected


def test_suspend():
    plan = [Msg("null")]

    processed_plan = list(suspend_wrapper(plan, 1))

    expected = [Msg("install_suspender", None, 1), Msg("null"), Msg("remove_suspender", None, 1)]

    assert processed_plan == expected


def test_md():
    def plan():
        yield from open_run(md={"a": 1})

    processed_plan = list(inject_md_wrapper(plan(), {"b": 2}))

    expected = [Msg("open_run", None, a=1, b=2)]

    assert processed_plan == expected


def test_finalize(hw):
    det = hw.det

    def plan():
        yield from [Msg("null")]

    def cleanup_plan():
        yield from [Msg("read", det)]

    # wrapper accepts list
    processed_plan = list(finalize_wrapper(plan(), [Msg("read", det)]))
    expected = [Msg("null"), Msg("read", det)]
    assert processed_plan == expected

    # or func that returns list
    def plan():
        yield from [Msg("null")]

    processed_plan = list(finalize_wrapper(plan(), lambda: [Msg("read", det)]))
    expected = [Msg("null"), Msg("read", det)]
    assert processed_plan == expected

    # or generator instance
    def plan():
        yield from [Msg("null")]

    processed_plan = list(finalize_wrapper(plan(), cleanup_plan()))
    expected = [Msg("null"), Msg("read", det)]
    assert processed_plan == expected

    # or generator func
    def plan():
        yield from [Msg("null")]

    processed_plan = list(finalize_wrapper(plan(), cleanup_plan))
    expected = [Msg("null"), Msg("read", det)]
    assert processed_plan == expected

    # decorator accepts generator func
    processed_plan = list(finalize_decorator(cleanup_plan)(plan)())
    expected = [Msg("null"), Msg("read", det)]
    assert processed_plan == expected

    # or func that returns list
    processed_plan = list(finalize_decorator(lambda: [Msg("read", det)])(plan)())
    expected = [Msg("null"), Msg("read", det)]
    assert processed_plan == expected

    # decorator does NOT accept list
    with pytest.raises(TypeError):
        list(finalize_decorator([Msg("read", det)])(plan)())

    # nor generator instance
    with pytest.raises(TypeError):
        list(finalize_decorator(cleanup_plan())(plan)())


def test_finalize_runs_after_error(RE, hw):
    det = hw.det

    def plan():
        yield from [Msg("null")]
        raise Exception

    msgs = []

    def accumulator(msg):
        msgs.append(msg)

    RE.msg_hook = accumulator
    try:
        RE(finalize_wrapper(plan(), [Msg("read", det)]))
    except Exception:
        pass  # swallow the Exception; we are interested in msgs below

    expected = [Msg("null"), Msg("read", det)]

    assert msgs == expected


def test_reset_positions(RE, hw):
    motor = hw.motor
    motor.set(5)

    msgs = []

    def accumulator(msg):
        msgs.append(msg)

    RE.msg_hook = accumulator

    def plan():
        yield from (m for m in [Msg("set", motor, 8)])

    RE(reset_positions_wrapper(plan()))

    expected = [Msg("set", motor, 8), Msg("set", motor, 5), Msg("wait")]

    for msg in msgs:
        msg.kwargs.pop("group", None)

    assert msgs == expected


def test_reset_positions_no_position_attr(RE, hw):
    motor = hw.motor_no_pos
    motor.set(5)

    msgs = []

    def accumulator(msg):
        msgs.append(msg)

    RE.msg_hook = accumulator

    def plan():
        yield from (m for m in [Msg("set", motor, 8)])

    RE(reset_positions_wrapper(plan()))

    expected = [Msg("read", motor), Msg("set", motor, 8), Msg("set", motor, 5), Msg("wait")]

    for msg in msgs:
        msg.kwargs.pop("group", None)

    assert msgs == expected


def test_relative_set(RE, hw):
    motor = hw.motor
    motor.set(5)

    msgs = []

    def accumulator(msg):
        msgs.append(msg)

    RE.msg_hook = accumulator

    def plan():
        yield from (m for m in [Msg("set", motor, 8)])

    RE(relative_set_wrapper(plan()))

    expected = [Msg("set", motor, 13)]

    for msg in msgs:
        msg.kwargs.pop("group", None)

    assert msgs == expected


def test_relative_set_no_position_attr(RE, hw):
    motor = hw.motor_no_pos
    motor.set(5)

    msgs = []

    def accumulator(msg):
        msgs.append(msg)

    RE.msg_hook = accumulator

    def plan():
        yield from (m for m in [Msg("set", motor, 8)])

    RE(relative_set_wrapper(plan()))

    expected = [Msg("read", motor), Msg("set", motor, 13)]

    for msg in msgs:
        msg.kwargs.pop("group", None)

    assert msgs == expected


def test_configure_count_time(RE, hw):
    det = hw.det_with_count_time
    det.count_time.put(3)

    def plan():
        yield from (m for m in [Msg("read", det)])

    msgs = []

    def accumulator(msg):
        msgs.append(msg)

    RE.msg_hook = accumulator

    RE(configure_count_time_wrapper(plan(), 7))

    expected = [
        Msg("set", det.count_time, 7),
        Msg("wait"),
        Msg("read", det),
        Msg("set", det.count_time, 3),
        Msg("wait"),
    ]

    for msg in msgs:
        msg.kwargs.pop("group", None)
        msg.kwargs.pop("timeout", None)

    assert msgs == expected


def test_repeater():
    def plan(*args):
        yield from args

    actual = list(repeater(3, plan, 1, 2, 3))
    assert actual == 3 * [1, 2, 3]

    p = repeater(None, plan, 1, 2, 3)
    assert next(p) == 1
    assert next(p) == 2
    assert next(p) == 3
    assert next(p) == 1


def test_caching_repeater():
    def plan(*args):
        yield from args

    plan_instance = plan(1, 2, 3)
    with pytest.warns(UserWarning):
        actual = list(caching_repeater(3, plan_instance))
    assert actual == 3 * [1, 2, 3]

    plan_instance = plan(1, 2, 3)
    p = caching_repeater(None, plan_instance)
    with pytest.warns(UserWarning):
        assert next(p) == 1
    assert next(p) == 2
    assert next(p) == 3
    assert next(p) == 1


def test_repeat(RE):
    # Check if lists and callables both work
    messages = [1, 2, 3]

    def plan():
        yield from messages

    num = 3
    expected = [Msg("checkpoint"), 1, 2, 3] * num
    assert list(repeat(plan, num=num)) == expected


def test_repeat_using_RE(RE):
    def plan():
        yield Msg("open_run")
        yield Msg("close_run")

    RE(repeat(plan, 2))


def test_trigger_and_read(hw):
    det = hw.det
    msgs = list(trigger_and_read([det]))
    expected = [
        Msg("trigger", det),
        Msg("wait", error_on_timeout=True, watch=()),
        Msg("create", name="primary"),
        Msg("read", det),
        Msg("save"),
    ]
    for msg in msgs:
        msg.kwargs.pop("group", None)
        msg.kwargs.pop("timeout", None)
    assert msgs == expected

    msgs = list(trigger_and_read([det], "custom"))
    expected = [
        Msg("trigger", det),
        Msg("wait", error_on_timeout=True, watch=()),
        Msg("create", name="custom"),
        Msg("read", det),
        Msg("save"),
    ]
    for msg in msgs:
        msg.kwargs.pop("group", None)
        msg.kwargs.pop("timeout", None)
    assert msgs == expected


def test_count_delay_argument(hw):
    # num=7 but delay only provides 5 entries
    with pytest.raises(ValueError):
        # count raises ValueError when delay generator is expired
        list(count([hw.det], num=7, delay=(2**i for i in range(5))))

    # num=6 with 5 delays between should product 6 readings
    msgs = count([hw.det], num=6, delay=(2**i for i in range(5)))
    read_count = len([msg for msg in msgs if msg.command == "read"])
    assert read_count == 6

    # num=5 with 5 delays should produce 5 readings
    msgs = count([hw.det], num=5, delay=(2**i for i in range(5)))
    read_count = len([msg for msg in msgs if msg.command == "read"])
    assert read_count == 5

    # num=4 with 5 delays should produce 4 readings
    msgs = count([hw.det], num=4, delay=(2**i for i in range(5)))
    read_count = len([msg for msg in msgs if msg.command == "read"])
    assert read_count == 4

    # num=None with 5 delays should produce 6 readings
    msgs = count([hw.det], num=None, delay=(2**i for i in range(5)))
    read_count = len([msg for msg in msgs if msg.command == "read"])
    assert read_count == 6


def test_plan_md(RE, hw):
    mutable = []
    md = {"color": "red"}

    def collector(name, doc):
        mutable.append(doc)

    # test genereator
    mutable.clear()
    RE(count([hw.det], md=md), collector)
    assert "color" in mutable[0]

    # test Plan with explicit __init__
    mutable.clear()
    RE(bp.count([hw.det], md=md), collector)
    assert "color" in mutable[0]

    # test Plan with implicit __init__ (created via metaclasss)
    mutable.clear()
    RE(bp.scan([hw.det], hw.motor, 1, 2, 2, md=md), collector)
    assert "color" in mutable[0]


def test_infinite_count(RE, hw):
    threading.Timer(1, RE.stop).start()
    docs = defaultdict(list)

    def collector(name, doc):
        docs[name].append(doc)

    with pytest.raises(RunEngineInterrupted):
        RE(count([hw.det], num=None), collector)

    assert len(docs["start"]) == 1
    assert len(docs["stop"]) == 1
    assert len(docs["descriptor"]) == 1
    assert len(docs["event"]) > 0


def test_no_rewind_device(hw):
    hw.det.rewindable = hw.bool_sig

    assert not all_safe_rewind([hw.det])


def test_monitor(RE, hw):
    from ophyd.sim import SynSignal

    signal = SynSignal(name="signal")
    signal.put(0.0)
    RE(monitor_during_wrapper(count([hw.det], 5), [signal]))


def test_per_step(RE, hw):
    # Check default behavior, using one motor and then two.
    RE(scan([hw.det], hw.motor, -1, 1, 3, per_step=one_nd_step))
    RE(scan([hw.det], hw.motor, -1, 1, hw.motor2, -1, 1, 3, per_step=one_nd_step))
    RE(inner_product_scan([hw.det], 3, hw.motor, -1, 1, per_step=one_nd_step))
    RE(inner_product_scan([hw.det], 3, hw.motor, -1, 1, hw.motor2, -1, 1, per_step=one_nd_step))

    # Check that scan still accepts old one_1d_step signature:
    RE(scan([hw.det], hw.motor, -1, 1, 3, per_step=one_1d_step))
    RE(rel_scan([hw.det], hw.motor, -1, 1, 3, per_step=one_1d_step))

    # Test that various error paths include a useful error message identifying
    # that the problem is with 'per_step':

    # You can't usage one_1d_step signature with more than one motor.
    with pytest.raises(TypeError) as excinfo:
        RE(scan([hw.det], hw.motor, -1, 1, hw.motor2, -1, 1, 3, per_step=one_1d_step))
    assert excinfo.match("Signature of per_step assumes 1D trajectory")

    # The signature must be either like one_1d_step or one_nd_step:
    def bad_sig(detectors, mtr, step): ...

    with pytest.raises(TypeError) as excinfo:
        RE(scan([hw.det], hw.motor, -1, 1, 3, per_step=bad_sig))
    assert excinfo.match("per_step must be a callable with the signature")


def test_stage_all_and_unstage_all(RE):
    from ophyd import StatusBase

    staged = {}
    unstaged = {}

    # Test support for old and new style devices
    class OldStyleDummy:
        def __init__(self, name: str) -> None:
            self.name: str = name

        def stage(self):
            staged[self.name] = True
            ttime.sleep(0.5)
            return [self]

        def unstage(self):
            unstaged[self.name] = True
            ttime.sleep(0.5)
            return [self]

    class NewStyleDummy:
        def __init__(self, name: str) -> None:
            self.name: str = name

        def _callback(self, d: dict, st: Status):
            d[self.name] = True
            st.set_finished()  # type: ignore

        def stage(self) -> Status:
            st = StatusBase()
            threading.Timer(0.5, self._callback, args=[staged, st]).start()
            return st

        def unstage(self) -> Status:
            st = StatusBase()
            threading.Timer(0.5, self._callback, args=[unstaged, st]).start()
            return st

    olddummy1 = OldStyleDummy("o1")
    olddummy2 = OldStyleDummy("o2")
    newdummy1 = NewStyleDummy("n1")
    newdummy2 = NewStyleDummy("n2")

    def plan():
        yield from stage_all(olddummy1, olddummy2, newdummy1, newdummy2)
        staged_keys = list(staged.keys())
        assert len(staged_keys) == 4
        assert staged_keys[:2] == ["o1", "o2"]
        assert "n1" in staged_keys[2:]
        assert "n2" in staged_keys[2:]

        yield from unstage_all(olddummy1, olddummy2, newdummy1, newdummy2)
        unstaged_keys = list(unstaged.keys())
        assert len(unstaged_keys) == 4
        assert unstaged_keys[:2] == ["o1", "o2"]
        assert "n1" in unstaged_keys[2:]
        assert "n2" in unstaged_keys[2:]

    start = ttime.monotonic()
    RE(plan())
    stop = ttime.monotonic()

    assert 3 < stop - start < 4


def test_old_style_wait(RE):
    class OldStyleDummy:
        def __init__(self, name: str) -> None:
            self.name: str = name

        def stage(self):
            ttime.sleep(0.5)
            return [self]

        def unstage(self):
            ttime.sleep(0.5)
            return [self]

    dummy = OldStyleDummy("o1")

    def stage_plan():
        yield from stage(dummy, wait=False)

    def unstage_plan():
        yield from unstage(dummy, wait=False)

    with pytest.raises(RuntimeError):
        RE(stage_plan())

    with pytest.raises(RuntimeError):
        RE(unstage_plan())


def test_custom_stream_name(RE, hw):
    def new_trigger_and_read(devices):
        return (yield from trigger_and_read(devices, name="secondary"))

    def new_per_step(detectors, motor, step):
        return (yield from one_1d_step(detectors, motor, step, take_reading=new_trigger_and_read))

    def new_per_shot(detectors):
        return (yield from one_shot(detectors, take_reading=new_trigger_and_read))

    RE(scan([hw.det], hw.motor, -1, 1, 3, per_step=new_per_step))
    RE(count([hw.det], 3, per_shot=new_per_shot))

    RE._require_stream_declaration = True
    with pytest.raises(IllegalMessageSequence):
        RE(scan([hw.det], hw.motor, -1, 1, 3, per_step=new_per_step))

    with pytest.raises(IllegalMessageSequence):
        RE(count([hw.det], 3, per_shot=new_per_shot))

    with pytest.raises(IllegalMessageSequence):
        RE(count([hw.det], 3, per_shot=one_shot))



import time as ttime
import warnings
from datetime import datetime

import numpy as np
import pytest
from event_model import RunRouter

import bluesky.plan_stubs as bps
import bluesky.preprocessors as bpp
from bluesky.callbacks.best_effort import BestEffortCallback
from bluesky.plans import grid_scan, scan
from bluesky.preprocessors import SupplementalData
from bluesky.tests.utils import DocCollector
from bluesky.utils import new_uid


def test_hints(RE, hw):
    motor = hw.motor
    expected_hint = {"fields": [motor.name]}
    assert motor.hints == expected_hint
    collector = []

    def collect(*args):
        collector.append(args)

    RE(scan([], motor, 1, 2, 2), {"descriptor": collect})
    name, doc = collector.pop()
    assert doc["hints"][motor.name] == expected_hint


def test_simple(RE, hw):
    bec = BestEffortCallback()
    RE.subscribe(bec)
    RE(scan([hw.ab_det], hw.motor, 1, 5, 5))


def test_disable(RE, hw):
    det, motor = hw.ab_det, hw.motor
    bec = BestEffortCallback()
    RE.subscribe(bec)

    bec.disable_table()

    RE(scan([det], motor, 1, 5, 5))
    assert bec._table is None

    bec.enable_table()

    RE(scan([det], motor, 1, 5, 5))
    assert bec._table is not None

    bec.peaks.com  # noqa: B018
    assert "det_a" in bec.peaks["com"]

    bec.clear()
    assert bec._table is None

    # smoke test
    bec.disable_plots()
    bec.enable_plots()
    bec.disable_baseline()
    bec.enable_baseline()
    bec.disable_heading()
    bec.enable_heading()


def test_blank_hints(RE, hw):
    bec = BestEffortCallback()
    RE.subscribe(bec)
    RE(scan([hw.ab_det], hw.motor, 1, 5, 5, md={"hints": {}}))


def test_with_baseline(RE, hw):
    bec = BestEffortCallback()
    RE.subscribe(bec)
    sd = SupplementalData(baseline=[hw.det])
    RE.preprocessors.append(sd)
    RE(scan([hw.ab_det], hw.motor, 1, 5, 5))


def test_underhinted_plan(RE, hw):
    bec = BestEffortCallback()
    RE.subscribe(bec)

    @bpp.run_decorator()
    def broken_plan(dets):
        yield from bps.declare_stream(*dets, name="primary")
        yield from bps.trigger_and_read(dets, name="primary")

    RE(broken_plan([hw.det]))


def test_live_grid(RE, hw):
    bec = BestEffortCallback()
    RE.subscribe(bec)
    RE(grid_scan([hw.det4], hw.motor1, 0, 1, 1, hw.motor2, 0, 1, 2, True))


def test_many_grids(RE, hw):
    bec = BestEffortCallback()
    RE.subscribe(bec)
    RE(grid_scan([hw.det1, hw.det2, hw.det3, hw.det4, hw.det5], hw.motor1, 0, 1, 1, hw.motor2, 0, 1, 2, True))
    # Exactly 3 Live plots should have x tick labels in a 3 column grid
    assert sum([bool(lg.ax.get_xticklabels()) for lg in list(bec._live_grids.values())[0].values()]) == 3
    # Exactly 5 axes should be visible of the 6 in the figure (ignoring color bars)
    assert (
        sum(
            [
                bool(ax.get_visible())
                for ax in list(list(bec._live_grids.values())[0].values())[0].ax.figure.axes
                if ax.get_label() != "<colorbar>"
            ]
        )
        == 5
    )


def test_push_start_document(capsys):
    """Pass the start document to BEC and verify if the scan information is printed correctly"""

    bec = BestEffortCallback()

    uid = new_uid()
    time = ttime.time()
    scan_id = 113435  # Just some arbitrary number

    # Include minimum information needed to print the header
    bec("start", {"scan_id": scan_id, "time": time, "uid": uid})

    captured = capsys.readouterr()
    assert f"Transient Scan ID: {scan_id}" in captured.out, "BestEffortCallback: Scan ID is not printed correctly"

    tt = datetime.fromtimestamp(time).utctimetuple()
    assert f"Time: {ttime.strftime('%Y-%m-%d %H:%M:%S', tt)}" in captured.out, (
        "BestEffortCallback: Scan time is not printed correctly"
    )
    assert f"Persistent Unique Scan ID: '{uid}'" in captured.out, (
        "BestEffortCallback: Scan UID is not printed correctly"
    )


def test_multirun_nested_plan(capsys, caplog, RE, hw):
    # This test only checks if the plan runs without crashing. If BEC crashes,
    #   the plan will still run, but data will not be displayed.
    @bpp.set_run_key_decorator(run="inner_run")
    def plan_inner():
        yield from grid_scan([hw.det4], hw.motor1, 0, 1, 1, hw.motor2, 0, 1, 2, True)

    def sequence():
        for n in range(5):
            yield from bps.mov(hw.motor, n * 0.1 + 1)
            yield from bps.trigger_and_read([hw.det1])

    @bpp.set_run_key_decorator(run="outer_run")
    @bpp.stage_decorator([hw.det1, hw.motor])
    @bpp.run_decorator(md={})
    def plan_outer():
        yield from bps.declare_stream(hw.det1, name="primary")

        yield from sequence()
        # Call inner plan from within the plan
        yield from plan_inner()
        # Run another set of commands
        yield from sequence()

    # The first test should fail. We check if expected error message is printed in case
    #   of failure.
    bec = BestEffortCallback()
    bec_token = RE.subscribe(bec)
    RE(plan_outer())

    captured = capsys.readouterr()

    # Check for the number of runs (the number of times UID is printed in the output)
    scan_uid_substr = "Persistent Unique Scan ID"
    n_runs = captured.out.count(scan_uid_substr)
    assert n_runs == 2, "scan output contains incorrect number of runs"
    # Check if the expected error message is printed once the callback fails. The same
    #   substring will be used in the second part of the test to check if BEC did not fail.
    err_msg_substr = "is being suppressed to not interrupt plan execution"
    assert err_msg_substr in str(caplog.text), (
        "Best Effort Callback failed, but expected error message was not printed"
    )

    RE.unsubscribe(bec_token)
    caplog.clear()

    # The second test should succeed, i.e. the error message should not be printed
    def factory(name, doc):
        bec = BestEffortCallback()
        return [bec], []

    rr = RunRouter([factory])
    RE.subscribe(rr)
    RE(plan_outer())

    captured = capsys.readouterr()
    n_runs = captured.out.count(scan_uid_substr)
    assert n_runs == 2, "scan output contains incorrect number of runs"
    assert err_msg_substr not in caplog.text, "Best Effort Callback failed while executing nested plans"


def test_plot_ints(RE):
    from ophyd import Signal

    import bluesky.plan_stubs as bps
    from bluesky.callbacks.best_effort import BestEffortCallback
    from bluesky.plans import count

    bec = BestEffortCallback()
    RE.subscribe(bec)

    s = Signal(name="s")
    RE(bps.mov(s, 0))
    assert s.describe()["s"]["dtype"] == "integer"
    s.kind = "hinted"

    with warnings.catch_warnings():
        warnings.simplefilter("error")
        RE(count([s], num=35))


def test_plot_prune_fifo(RE, hw):
    bec = BestEffortCallback()
    RE.subscribe(bec)

    num_pruned = 2

    # create the LivePlot
    RE(bps.repeater(num_pruned, scan, [hw.ab_det], hw.motor, 1, 5, 5))

    # test it
    assert len(bec._live_plots) == 1

    # get the reference key for our LivePlot dict
    uuid = next(iter(bec._live_plots))
    assert len(bec._live_plots[uuid]) == 1

    # get reference key for our detector
    det_name = next(iter(bec._live_plots[uuid]))
    # should be same as hw.ab_det.a.name (`.a` comes from .read_attrs[0]), prove it now
    assert det_name == hw.ab_det.a.name

    # get the LivePlot object
    lp = bec._live_plots[uuid][det_name]

    assert lp is not None
    assert len(lp.ax.lines) == num_pruned

    # prune the LivePlot (has no effect since we have exact number to keep)
    bec.plot_prune_fifo(num_pruned, hw.motor, hw.ab_det.a)
    assert len(lp.ax.lines) == num_pruned

    # add more lines to the LivePlot
    RE(bps.repeater(num_pruned, scan, [hw.ab_det], hw.motor, 1, 5, 5))

    # get the LivePlot object, again, in case the UUID was changed
    assert len(bec._live_plots) == 1
    uuid = next(iter(bec._live_plots))
    lp = bec._live_plots[uuid][det_name]
    assert lp is not None
    assert len(lp.ax.lines) == num_pruned * 2

    # prune again, this time reduces number of lines
    bec.plot_prune_fifo(num_pruned, hw.motor, hw.ab_det.a)
    assert len(lp.ax.lines) == num_pruned


def test_bec_peak_stats_derivative_and_stats(RE, hw):
    bec = BestEffortCallback(calc_derivative_and_stats=True)
    RE.subscribe(bec)

    c = DocCollector()
    RE.subscribe(c.insert)

    res = RE(scan([hw.ab_det], hw.motor, 1, 5, 5))

    if RE.call_returns_result:
        uid = res.run_start_uids[0]
    else:
        uid = res[0]

    desc_uid = c.descriptor[uid][0]["uid"]
    ps = bec._peak_stats[desc_uid]["det_a"]

    assert hasattr(ps, "derivative_stats")

    fields = ["min", "max", "com", "cen", "fwhm", "crossings"]
    der_fields = ["x", "y"] + fields
    for field in der_fields:
        assert hasattr(ps.derivative_stats, field), f"{field} is not an attribute of ps.der"

    assert isinstance(ps.__repr__(), str)

    # These imports are needed by the `eval` below:
    from numpy import array  # noqa F401
    from collections import OrderedDict  # noqa F401

    out = eval(str(ps))
    assert isinstance(out, dict)
    for key in ("stats", "derivative_stats"):
        assert key in out

    for field in fields:
        stats_value = getattr(ps.stats, field)
        out_value = out["stats"][field]
        if stats_value is not None:
            assert np.allclose(stats_value, out_value)
        else:
            stats_value == out_value  # noqa: B015

    for field in der_fields:
        stats_value = getattr(ps.derivative_stats, field)
        out_value = out["derivative_stats"][field]
        if stats_value is not None:
            assert np.allclose(stats_value, out_value)
        else:
            stats_value == out_value  # noqa: B015


def test_many_motors(RE, hw):
    """Ensure appropriate behavior for too many motors to plot. No figures with warning, and a table."""
    dets = [hw.ab_det]
    motors = [hw.motor, hw.motor1, hw.motor2, hw.motor3]
    bec = BestEffortCallback()
    RE.subscribe(bec)
    movement = [(motor, 1, 5, 5) for motor in motors]
    with pytest.warns((RuntimeWarning, UserWarning)):
        RE(grid_scan(dets, *[item for sublist in movement for item in sublist]))

    assert not bec._live_plots
    assert not bec._live_grids
    assert not bec._live_scatters
    assert bec._table is not None



from math import ceil

import pytest

from bluesky.consolidators import HDF5Consolidator, consolidator_factory


@pytest.fixture
def descriptor():
    return {
        "data_keys": {
            "test_img": {
                "shape": [1, 10, 15],
                "dtype": "array",
                "dtype_numpy": "<f8",
                "external": "STREAM:",
                "object_name": "test_object",
            },
            "test_7_imgs": {
                "shape": [7, 10, 15],
                "dtype": "array",
                "dtype_numpy": "<f8",
                "external": "STREAM:",
                "object_name": "test_object",
            },
            "test_6_imgs": {
                "shape": [6, 10, 15],
                "dtype": "array",
                "dtype_numpy": "<f8",
                "external": "STREAM:",
                "object_name": "test_object",
            },
            "test_cube": {
                "shape": [1, 10, 15, 3],
                "dtype": "array",
                "dtype_numpy": "<f8",
                "external": "STREAM:",
                "object_name": "test_object",
            },
            "test_7_cubes": {
                "shape": [7, 10, 15, 3],
                "dtype": "array",
                "dtype_numpy": "<f8",
                "external": "STREAM:",
                "object_name": "test_object",
            },
            "test_arr": {
                "shape": [
                    1,
                    3,
                ],
                "dtype": "array",
                "dtype_numpy": "<f8",
                "external": "STREAM:",
                "object_name": "test_object",
            },
            "test_7_arrs": {
                "shape": [
                    7,
                    3,
                ],
                "dtype": "array",
                "dtype_numpy": "<f8",
                "external": "STREAM:",
                "object_name": "test_object",
            },
            "test_num": {
                "shape": [
                    1,
                ],
                "dtype": "number",
                "dtype_numpy": "<f8",
                "external": "STREAM:",
                "object_name": "test_object",
            },
            "test_7_nums": {
                "shape": [
                    7,
                ],
                "dtype": "number",
                "dtype_numpy": "<f8",
                "external": "STREAM:",
                "object_name": "test_object",
            },
        },
        "uid": "descriptor-uid",
    }


@pytest.fixture
def hdf5_stream_resource_factory():
    return lambda data_key, chunk_shape: {
        "data_key": data_key,
        "mimetype": "application/x-hdf5",
        "uri": "file://localhost/test/file/path",
        "resource_path": "test_file.h5",
        "parameters": {
            "dataset": f"entry/data/{data_key}",
            "swmr": True,
            "chunk_shape": chunk_shape,
        },
        "uid": f"stream-resource-uid-{data_key}",
    }


@pytest.fixture
def image_seq_stream_resource_factory():
    format_to_mimetype = {
        "jpeg": "image/jpeg",
        "jpg": "image/jpeg",
        "tiff": "image/tiff",
        "tif": "image/tiff",
    }
    return lambda image_format, data_key, chunk_shape: {
        "data_key": data_key,
        "mimetype": f"multipart/related;type={format_to_mimetype[image_format]}",
        "uri": "file://localhost/test/file/path",
        "parameters": {"chunk_shape": chunk_shape, "template": "img_{:06d}." + image_format},
        "uid": f"stream-resource-uid-{data_key}",
    }


@pytest.fixture
def csv_stream_resource_factory():
    return lambda data_key, chunk_shape: {
        "data_key": data_key,
        "mimetype": "text/csv;header=absent",
        "uri": "file://localhost/test/file/path",
        "resource_path": "test_file.csv",
        "parameters": {"chunk_shape": chunk_shape},
        "uid": f"stream-resource-uid-{data_key}",
    }


@pytest.fixture
def stream_datum_factory():
    return lambda data_key, indx, i_start, i_stop: {
        "seq_nums": {"start": i_start + 1, "stop": i_stop + 1},
        "indices": {"start": i_start, "stop": i_stop},
        "descriptor": "descriptor-uid",
        "stream_resource": f"stream-resource-uid-{data_key}",
        "uid": f"stream-datum-uid-{data_key}/{indx}",
    }


# Tuples of (data_key, frames_per_datum, join_method, expected_shape)
shape_testdata = [
    # 5 events, 1 or 7 image per event, 10x15 pixels
    ("test_img", 1, "concat", (5, 10, 15)),
    ("test_7_imgs", 7, "concat", (35, 10, 15)),
    ("test_img", 1, "stack", (5, 1, 10, 15)),
    ("test_7_imgs", 7, "stack", (5, 7, 10, 15)),
    # 5 events, 1 or 7 cube per event, 10x15x3 pixels
    ("test_cube", 1, "concat", (5, 10, 15, 3)),
    ("test_7_cubes", 7, "concat", (35, 10, 15, 3)),
    ("test_cube", 1, "stack", (5, 1, 10, 15, 3)),
    ("test_7_cubes", 7, "stack", (5, 7, 10, 15, 3)),
    # 5 events, 1 or 7 array per event, 1 element in array
    ("test_arr", 1, "concat", (5, 3)),
    ("test_7_arrs", 7, "concat", (35, 3)),
    ("test_arr", 1, "stack", (5, 1, 3)),
    ("test_7_arrs", 7, "stack", (5, 7, 3)),
    # 5 events, 1 or 7 number per event
    ("test_num", 1, "concat", (5,)),
    ("test_7_nums", 7, "concat", (35,)),
    ("test_num", 1, "stack", (5, 1)),
    ("test_7_nums", 7, "stack", (5, 7)),
]


@pytest.mark.parametrize("data_key, frames_per_datum, join_method, expected", shape_testdata)
def test_hdf5_shape(
    descriptor,
    hdf5_stream_resource_factory,
    stream_datum_factory,
    data_key,
    frames_per_datum,
    join_method,
    expected,
):
    stream_resource = hdf5_stream_resource_factory(data_key=data_key, chunk_shape=())
    cons = HDF5Consolidator(stream_resource, descriptor)
    cons.join_method = join_method
    assert cons.shape == (0, *expected[1:])
    for i in range(5):
        doc = stream_datum_factory(data_key, i, i, i + 1)
        cons.consume_stream_datum(doc)
    assert cons.shape == expected


supported_image_seq_formats = ["jpeg", "tiff", "jpg", "tif"]


@pytest.mark.parametrize("data_key, frames_per_datum, join_method, expected", shape_testdata)
@pytest.mark.parametrize("image_format", supported_image_seq_formats)
@pytest.mark.parametrize("indx_per_stream_datum_doc", [1, 2, 3, 5])
def test_tiff_and_jpeg_shape(
    descriptor,
    image_seq_stream_resource_factory,
    stream_datum_factory,
    image_format,
    data_key,
    frames_per_datum,
    join_method,
    expected,
    indx_per_stream_datum_doc,
):
    stream_resource = image_seq_stream_resource_factory(
        image_format=image_format, data_key=data_key, chunk_shape=(1,)
    )
    cons = consolidator_factory(stream_resource, descriptor)
    cons.join_method = join_method
    assert cons.shape == (0, *expected[1:])
    for i in range(ceil(5 / indx_per_stream_datum_doc)):
        doc = stream_datum_factory(
            data_key, i, i * indx_per_stream_datum_doc, min((i + 1) * indx_per_stream_datum_doc, 5)
        )
        cons.consume_stream_datum(doc)
    assert cons.shape == expected

    # Stackable case here corresponds to multipage tiffs (AD does not support them though)
    assert len(cons.assets) == 5 * frames_per_datum if join_method == "concat" else 5


# Tuples of (data_key, chunk_shape, expected_shape, expected_chunks)
csv_testdata = [
    # 5 events, 1 or 7 array per event, 1 element in array
    ("test_arr", (1,), (5, 3), ((1,) * 5, (3,))),
    ("test_7_arrs", (1,), (35, 3), ((1,) * 35, (3,))),
    ("test_arr", (), (5, 3), ((5,), (3,))),
    ("test_7_arrs", (), (35, 3), ((35,), (3,))),
    ("test_arr", (10,), (5, 3), ((1,) * 5, (3,))),
    ("test_7_arrs", (10,), (35, 3), ((7, 7, 7, 7, 7), (3,))),
    ("test_7_arrs", (3,), (35, 3), ((3, 3, 1) * 5, (3,))),
]


@pytest.mark.parametrize("data_key, chunk_shape, expected_shape, expected_chunks", csv_testdata)
def test_csv_shape_and_chunks(
    descriptor,
    csv_stream_resource_factory,
    stream_datum_factory,
    data_key,
    chunk_shape,
    expected_shape,
    expected_chunks,
):
    stream_resource = csv_stream_resource_factory(data_key=data_key, chunk_shape=chunk_shape)
    cons = consolidator_factory(stream_resource, descriptor)
    assert cons.join_method == "concat"
    assert not cons.join_chunks
    assert cons.shape == (0, *expected_shape[1:])
    for i in range(5):
        doc = stream_datum_factory(data_key, i, i, i + 1)
        cons.consume_stream_datum(doc)
    assert cons.shape == expected_shape
    assert cons.chunks == expected_chunks


# Tuples of (data_key, join_method, join_chunks, chunk_shape, expected_chunks)
chunk_hdf5_testdata = [
    ("test_img", "stack", True, (), ((5,), (1,), (10,), (15,))),
    ("test_img", "stack", True, (1, 1, 10, 15), ((1, 1, 1, 1, 1), (1,), (10,), (15,))),
    ("test_img", "stack", True, (2,), ((2, 2, 1), (1,), (10,), (15,))),
    ("test_img", "stack", True, (5, 1, 10, 15), ((5,), (1,), (10,), (15,))),
    ("test_img", "stack", True, (10, 1), ((5,), (1,), (10,), (15,))),
    ("test_img", "stack", True, (3, 1, 4, 5), ((3, 2), (1,), (4, 4, 2), (5, 5, 5))),
    ("test_7_imgs", "stack", True, (), ((5,), (7,), (10,), (15,))),
    ("test_7_imgs", "stack", True, (1, 1), ((1, 1, 1, 1, 1), (1, 1, 1, 1, 1, 1, 1), (10,), (15,))),
    ("test_7_imgs", "stack", True, (2,), ((2, 2, 1), (7,), (10,), (15,))),
    ("test_7_imgs", "stack", True, (5, 1, 10, 15), ((5,), (1, 1, 1, 1, 1, 1, 1), (10,), (15,))),
    ("test_7_imgs", "stack", True, (10, 5), ((5,), (5, 2), (10,), (15,))),
    (
        "test_7_imgs",
        "stack",
        True,
        (3, 4, 5, 6),
        (
            (3, 2),
            (4, 3),
            (5, 5),
            (6, 6, 3),
        ),
    ),
    ("test_cube", "stack", True, (3, 1, 4, 5, 3), ((3, 2), (1,), (4, 4, 2), (5, 5, 5), (3,))),
    ("test_7_cubes", "stack", True, (3, 4, 5, 6, 7), ((3, 2), (4, 3), (5, 5), (6, 6, 3), (3,))),
    ("test_arr", "stack", True, (5, 1, 1), ((5,), (1,), (1, 1, 1))),
    ("test_arr", "stack", True, (2,), ((2, 2, 1), (1,), (3,))),
    ("test_7_arrs", "stack", True, (5, 1, 1), ((5,), (1, 1, 1, 1, 1, 1, 1), (1, 1, 1))),
    ("test_7_arrs", "stack", True, (2,), ((2, 2, 1), (7,), (3,))),
    ("test_num", "stack", True, (), ((5,), (1,))),
    ("test_num", "stack", True, (2,), ((2, 2, 1), (1,))),
    ("test_7_nums", "stack", True, (), ((5,), (7,))),
    ("test_7_nums", "stack", True, (2,), ((2, 2, 1), (7,))),
    ("test_7_nums", "stack", True, (2, 3), ((2, 2, 1), (3, 3, 1))),
    ("test_img", "concat", True, (), ((5,), (10,), (15,))),
    ("test_img", "concat", True, (1, 10, 15), ((1, 1, 1, 1, 1), (10,), (15,))),
    ("test_img", "concat", True, (2,), ((2, 2, 1), (10,), (15,))),
    ("test_img", "concat", True, (5, 10, 15), ((5,), (10,), (15,))),
    ("test_img", "concat", True, (10, 1), ((5,), (1,) * 10, (15,))),
    ("test_img", "concat", True, (3, 4, 5), ((3, 2), (4, 4, 2), (5, 5, 5))),
    ("test_7_imgs", "concat", True, (), ((35,), (10,), (15,))),
    ("test_7_imgs", "concat", False, (), ((35,), (10,), (15,))),
    ("test_7_imgs", "concat", True, (1, 1), ((1,) * 35, (1,) * 10, (15,))),
    ("test_7_imgs", "concat", False, (2,), ((2, 2, 2, 1) * 5, (10,), (15,))),
    ("test_7_imgs", "concat", True, (2,), ((2,) * 17 + (1,), (10,), (15,))),
    ("test_7_imgs", "concat", True, (5, 10, 15), ((5,) * 7, (10,), (15,))),
    ("test_7_imgs", "concat", False, (5, 10, 15), ((5, 2) * 5, (10,), (15,))),
    ("test_7_imgs", "concat", True, (10, 5), ((10, 10, 10, 5), (5, 5), (15,))),
    ("test_7_imgs", "concat", False, (10, 5), ((7,) * 5, (5, 5), (15,))),
    (
        "test_7_imgs",
        "concat",
        True,
        (3, 5, 6),
        (
            (3,) * 11 + (2,),
            (5, 5),
            (6, 6, 3),
        ),
    ),
    (
        "test_7_imgs",
        "concat",
        False,
        (3, 5, 6),
        (
            (3, 3, 1) * 5,
            (5, 5),
            (6, 6, 3),
        ),
    ),
    ("test_cube", "concat", True, (3, 4, 5, 3), ((3, 2), (4, 4, 2), (5, 5, 5), (3,))),
    ("test_7_cubes", "concat", True, (3, 5, 6, 7), ((3,) * 11 + (2,), (5, 5), (6, 6, 3), (3,))),
    ("test_7_cubes", "concat", False, (3, 5, 6, 7), ((3, 3, 1) * 5, (5, 5), (6, 6, 3), (3,))),
    ("test_arr", "concat", True, (5, 1), ((5,), (1, 1, 1))),
    ("test_arr", "concat", True, (2,), ((2, 2, 1), (3,))),
    ("test_7_arrs", "concat", True, (5, 1), ((5,) * 7, (1, 1, 1))),
    ("test_7_arrs", "concat", False, (5, 1), ((5, 2) * 5, (1, 1, 1))),
    ("test_7_arrs", "concat", True, (2,), ((2,) * 17 + (1,), (3,))),
    ("test_7_arrs", "concat", False, (2,), ((2, 2, 2, 1) * 5, (3,))),
    ("test_num", "concat", True, (), ((5,),)),
    ("test_num", "concat", True, (2,), ((2, 2, 1),)),
    ("test_7_nums", "concat", True, (), ((35,),)),
    ("test_7_nums", "concat", False, (), ((35,),)),
    ("test_7_nums", "concat", True, (3,), ((3,) * 11 + (2,),)),
    ("test_7_nums", "concat", False, (3,), ((3, 3, 1) * 5,)),
]


@pytest.mark.parametrize("data_key, join_method, join_chunks, chunk_shape, expected", chunk_hdf5_testdata)
def test_hdf5_chunks(
    descriptor,
    hdf5_stream_resource_factory,
    stream_datum_factory,
    data_key,
    join_method,
    join_chunks,
    chunk_shape,
    expected,
):
    stream_resource = hdf5_stream_resource_factory(data_key=data_key, chunk_shape=chunk_shape)
    cons = HDF5Consolidator(stream_resource, descriptor)
    cons.join_method = join_method
    cons.join_chunks = join_chunks
    assert cons.chunks == ((0,), *expected[1:])
    for i in range(5):
        doc = stream_datum_factory(data_key, i, i, i + 1)
        cons.consume_stream_datum(doc)
    assert cons.chunks == expected


# Tuples of (data_key, join_method, join_chunks, frames_per_datum, indx_per_stream_datum_doc, chunk_shape, expected_chunks) # noqa
chunk_tiff_testdata = [
    ("test_img", "stack", True, 1, 1, (1,), ((1, 1, 1, 1, 1), (1,), (10,), (15,))),
    ("test_img", "stack", True, 1, 2, (1,), ((1, 1, 1, 1, 1), (1,), (10,), (15,))),
    ("test_img", "stack", True, 1, 5, (1,), ((1, 1, 1, 1, 1), (1,), (10,), (15,))),
    ("test_6_imgs", "stack", True, 6, 1, (1,), ((1,) * 5, (6,), (10,), (15,))),
    ("test_6_imgs", "concat", True, 6, 1, (1,), ((1,) * 30, (10,), (15,))),
    ("test_6_imgs", "stack", True, 6, 2, (2,), ((2, 2, 1), (6,), (10,), (15,))),
    ("test_6_imgs", "concat", True, 6, 2, (2,), ((2,) * 15, (10,), (15,))),
    ("test_6_imgs", "stack", True, 6, 4, (3,), ((3, 2), (6,), (10,), (15,))),
    ("test_6_imgs", "concat", True, 6, 4, (3,), ((3,) * 10, (10,), (15,))),
    ("test_6_imgs", "stack", True, 6, 1, (5,), None),  # chunk_shape[0] must devide the number of frames
    ("test_6_imgs", "concat", True, 6, 1, (5,), None),
    ("test_6_imgs", "concat", True, 6, 10, (10,), None),
]


@pytest.mark.parametrize(
    "data_key, join_method, join_chunks, frames_per_datum, indx_per_stream_datum_doc, chunk_shape, expected_chunks",  # noqa
    chunk_tiff_testdata,
)
@pytest.mark.parametrize("image_format", supported_image_seq_formats)
def test_tiff_and_jpeg_chunks(
    descriptor,
    image_seq_stream_resource_factory,
    stream_datum_factory,
    image_format,
    data_key,
    join_method,
    join_chunks,
    frames_per_datum,
    indx_per_stream_datum_doc,
    chunk_shape,
    expected_chunks,
):
    """Test the chunking of (possibly multipage) tiff and jpeg datasets and the number of registered files."""

    stream_resource = image_seq_stream_resource_factory(
        image_format=image_format, data_key=data_key, chunk_shape=chunk_shape
    )
    if expected_chunks is None:
        with pytest.raises(AssertionError):
            cons = consolidator_factory(stream_resource, descriptor)
        return

    cons = consolidator_factory(stream_resource, descriptor)
    cons.join_method = join_method
    cons.join_chunks = join_chunks
    assert cons.chunks == ((0,), *expected_chunks[1:])
    for i in range(ceil(5 / indx_per_stream_datum_doc)):
        doc = stream_datum_factory(
            data_key, i, i * indx_per_stream_datum_doc, min((i + 1) * indx_per_stream_datum_doc, 5)
        )
        cons.consume_stream_datum(doc)
    assert cons.chunks == expected_chunks

    # Check the number of registered files
    assert len(cons.assets) == 5 * frames_per_datum / expected_chunks[0][0] if join_method == "concat" else 5


# Tuples of (filename, original_template, expected_template, formatted)
template_testdata = [
    ("", "img_{:06d}", "img_{:06d}", "img_000042"),
    ("img", "{:s}_{:06d}", "img_{:06d}", "img_000042"),
    ("img", "%s_%06d", "img_{:06d}", "img_000042"),
    ("", "img%s_%06d", "img_{:06d}", "img_000042"),
    ("img", "%s_%1d", "img_{:1d}", "img_42"),
    ("img", "%s_%-6d", "img_{:<6d}", "img_42    "),
    ("img", "%s_%+06d", "img_{:+06d}", "img_+00042"),
    ("img", "%s_% 06d", "img_{: 06d}", "img_ 00042"),
    ("img", "%s_%-+6d", "img_{:<+6d}", "img_+42   "),
    ("img", "%s_%- 6d", "img_{:< 6d}", "img_ 42   "),
    ("img", "%s_%6.6d", "img_{:06d}", "img_000042"),
]


@pytest.mark.parametrize("image_format", supported_image_seq_formats)
@pytest.mark.parametrize("filename, original_template, expected_template, formatted", template_testdata)
def test_name_templating(
    descriptor,
    image_seq_stream_resource_factory,
    image_format,
    filename,
    original_template,
    expected_template,
    formatted,
):
    stream_resource = image_seq_stream_resource_factory(
        image_format=image_format, data_key="test_img", chunk_shape=((1),)
    )
    stream_resource["parameters"]["template"] = f"{original_template}.{image_format}"
    if filename:
        stream_resource["parameters"]["filename"] = filename
    cons = consolidator_factory(stream_resource, descriptor)
    assert cons.template == f"{expected_template}.{image_format}"
    assert cons.template.format(42) == f"{formatted}.{image_format}"



"""
The key classes needed to use 0MQ for multiprocess document communication.

`Publisher` : subscribe this to the RE to emit the documents.  Expects a server to
have a SUBSCRIBE port open to PUB to.

`RemoteDispatcher` : subscribe callbacks to this class in a remote process.  Expects
a server to have a PUB port open to SUBSCRIBE to.

`Proxy` : server that binds ports for Pubslisher to push to and the Dispatcher
to listen to.  Typically this is started with the cli tool ``bluesky-zmq-proxy``

"""

import asyncio
import copy
import pickle
import warnings

from ..run_engine import Dispatcher, DocumentNames


class Bluesky0MQDecodeError(Exception):
    """Custom exception class for things that go wrong reading message from wire."""

    ...


class Publisher:
    """
    A callback that publishes documents to a 0MQ proxy.

    Parameters
    ----------
    address : string or tuple
        Address of a running 0MQ proxy, given either as a string like
        ``'127.0.0.1:5567'`` or as a tuple like ``('127.0.0.1', 5567)``
    prefix : bytes, optional
        User-defined bytestring used to distinguish between multiple
        Publishers. May not contain b' '.
    RE : ``bluesky.RunEngine``, optional
        DEPRECATED.
        RunEngine to which the Publisher will be automatically subscribed
        (and, more importantly, unsubscribed when it is closed).
    zmq : object, optional
        By default, the 'zmq' module is imported and used. Anything else
        mocking its interface is accepted.
    serializer: function, optional
        optional function to serialize data. Default is pickle.dumps

    Examples
    --------

    Publish from a RunEngine to a Proxy running on localhost on port 5567.

    >>> publisher = Publisher('localhost:5567')
    >>> RE = RunEngine({})
    >>> RE.subscribe(publisher)
    """

    def __init__(self, address, *, prefix=b"", RE=None, zmq=None, serializer=pickle.dumps):
        if RE is not None:
            warnings.warn(  # noqa: B028
                "The RE argument to Publisher is deprecated and "
                "will be removed in a future release of bluesky. "
                "Update your code to subscribe this Publisher "
                "instance to (and, if needed, unsubscribe from) to "
                "the RunEngine manually."
            )
        if isinstance(prefix, str):
            raise ValueError("prefix must be bytes, not string")
        if b" " in prefix:
            raise ValueError(f"prefix {prefix!r} may not contain b' '")
        if zmq is None:
            import zmq
        if isinstance(address, str):
            address = address.split(":", maxsplit=1)
        self.address = (address[0], int(address[1]))
        self.RE = RE
        url = "tcp://%s:%d" % self.address
        self._prefix = bytes(prefix)
        self._context = zmq.Context()
        self._socket = self._context.socket(zmq.PUB)
        self._socket.connect(url)
        if RE:
            self._subscription_token = RE.subscribe(self)
        self._serializer = serializer

    def __call__(self, name, doc):
        doc = copy.deepcopy(doc)
        message = b" ".join([self._prefix, name.encode(), self._serializer(doc)])
        self._socket.send(message)

    def close(self):
        if self.RE:
            self.RE.unsubscribe(self._subscription_token)
        self._socket.close()
        self._context.destroy()  # close Socket(s); terminate Context


class Proxy:
    """
    Start a 0MQ proxy on the local host.

    Parameters
    ----------
    in_port : int, optional
        Port that RunEngines should broadcast to. If None, a random port is
        used.
    out_port : int, optional
        Port that subscribers should subscribe to. If None, a random port is
        used.
    zmq : object, optional
        By default, the 'zmq' module is imported and used. Anything else
        mocking its interface is accepted.

    Attributes
    ----------
    in_port : int
        Port that RunEngines should broadcast to.
    out_port : int
        Port that subscribers should subscribe to.
    closed : boolean
        True if the Proxy has already been started and subsequently
        interrupted and is therefore unusable.

    Examples
    --------

    Run on specific ports.

    >>> proxy = Proxy(5567, 5568)
    >>> proxy
    Proxy(in_port=5567, out_port=5568)
    >>> proxy.start()  # runs until interrupted

    Run on random ports, and access those ports before starting.

    >>> proxy = Proxy()
    >>> proxy
    Proxy(in_port=56504, out_port=56505)
    >>> proxy.in_port
    56504
    >>> proxy.out_port
    56505
    >>> proxy.start()  # runs until interrupted
    """

    def __init__(self, in_port=None, out_port=None, *, zmq=None):
        if zmq is None:
            import zmq
        self.zmq = zmq
        self.closed = False
        try:
            context = zmq.Context(1)
            # Socket facing clients
            frontend = context.socket(zmq.SUB)
            if in_port is None:
                in_port = frontend.bind_to_random_port("tcp://*")
            else:
                frontend.bind("tcp://*:%d" % in_port)

            frontend.setsockopt_string(zmq.SUBSCRIBE, "")

            # Socket facing services
            backend = context.socket(zmq.PUB)
            if out_port is None:
                out_port = backend.bind_to_random_port("tcp://*")
            else:
                backend.bind("tcp://*:%d" % out_port)
        except BaseException:
            # Clean up whichever components we have defined so far.
            try:
                frontend.close()
            except NameError:
                ...
            try:
                backend.close()
            except NameError:
                ...
            context.destroy()
            raise
        else:
            self.in_port = in_port
            self.out_port = out_port
            self._frontend = frontend
            self._backend = backend
            self._context = context

    def start(self):
        if self.closed:
            raise RuntimeError(
                f"This Proxy has already been started and interrupted. Create a fresh instance with {repr(self)}"
            )
        try:
            self.zmq.device(self.zmq.FORWARDER, self._frontend, self._backend)
        finally:
            self.closed = True
            self._frontend.close()
            self._backend.close()
            self._context.destroy()

    def __repr__(self):
        return "{}(in_port={in_port}, out_port={out_port})".format(type(self).__name__, **vars(self))


class RemoteDispatcher(Dispatcher):
    """
    Dispatch documents received over the network from a 0MQ proxy.

    Parameters
    ----------
    address : tuple
        Address of a running 0MQ proxy, given either as a string like
        ``'127.0.0.1:5567'`` or as a tuple like ``('127.0.0.1', 5567)``
    prefix : bytes, optional
        User-defined bytestring used to distinguish between multiple
        Publishers. If set, messages without this prefix will be ignored.
        If unset, no mesages will be ignored.
    loop : zmq.asyncio.ZMQEventLoop, optional
    zmq : object, optional
        By default, the 'zmq' module is imported and used. Anything else
        mocking its interface is accepted.
    zmq_asyncio : object, optional
        By default, the 'zmq.asyncio' module is imported and used. Anything
        else mocking its interface is accepted.
    deserializer: function, optional
        optional function to deserialize data. Default is pickle.loads

    Examples
    --------

    Print all documents generated by remote RunEngines.

    >>> d = RemoteDispatcher(('localhost', 5568))
    >>> d.subscribe(print)
    >>> d.start()  # runs until interrupted
    """

    def __init__(
        self,
        address,
        *,
        prefix=b"",
        loop=None,
        zmq=None,
        zmq_asyncio=None,
        deserializer=pickle.loads,
        strict=False,
    ):
        if isinstance(prefix, str):
            raise ValueError("prefix must be bytes, not string")
        if b" " in prefix:
            raise ValueError(f"prefix {prefix!r} may not contain b' '")
        self._prefix = prefix
        if zmq is None:
            import zmq
        if zmq_asyncio is None:
            import zmq.asyncio as zmq_asyncio
        if isinstance(address, str):
            address = address.split(":", maxsplit=1)
        self._deserializer = deserializer
        self.address = (address[0], int(address[1]))

        if loop is None:
            loop = asyncio.new_event_loop()
        self.loop = loop
        self._context = None
        self._socket = None

        def __finish_setup():
            asyncio.set_event_loop(self.loop)

            self._context = zmq_asyncio.Context()
            self._socket = self._context.socket(zmq.SUB)

            url = "tcp://%s:%d" % self.address
            self._socket.connect(url)
            self._socket.setsockopt_string(zmq.SUBSCRIBE, "")

        self.__factory = __finish_setup
        self._task = None
        self.closed = False
        self._strict = strict
        super().__init__()

    async def _poll(self):
        our_prefix = self._prefix  # local var to save an attribute lookup
        while True:
            message = await self._socket.recv()
            try:
                prefix, name, doc = message.split(b" ", 2)
            except ValueError as e:
                if self._strict:
                    raise Bluesky0MQDecodeError from e
                else:
                    print(
                        f"The message {message} could not be split into "
                        "three parts by b' '.  Dropping message on floor "
                        "and continuing"
                        f"\n\n{e}"
                    )
                    continue

            try:
                name = name.decode()
            except UnicodeDecodeError as e:
                if self._strict:
                    raise Bluesky0MQDecodeError from e
                else:
                    print(
                        f"The name {name} can not be decoded as utf-8. "
                        "Dropping message on the floor and continuing. "
                        f"\n\n{e}"
                    )
                    continue
            if (not our_prefix) or prefix == our_prefix:
                try:
                    doc = self._deserializer(doc)
                except Exception as e:
                    if self._strict:
                        raise Bluesky0MQDecodeError from e
                    else:
                        if len(doc) > 1024:
                            msg_doc = doc[:1024] + b"--SNIPPED--"
                        else:
                            msg_doc = doc
                        print(
                            f"Failed to deserialize the {name} document "
                            f"{msg_doc} using {self._deserializer}. "
                            "Dropping on floor and continuing"
                            f"\n\n{e}"
                        )
                        continue
                self.loop.call_soon(self.process, DocumentNames[name], doc)

    def start(self):
        if self.closed:
            raise RuntimeError(
                "This RemoteDispatcher has already been "
                "started and interrupted. Create a fresh "
                f"instance with {self!r}"
            )
        try:
            self.__factory()
            self._task = self.loop.create_task(self._poll())
            self.loop.run_until_complete(self._task)
            task_exception = self._task.exception()
            if task_exception is not None:
                raise task_exception
        finally:
            self.stop()

    def stop(self):
        if self._task is not None:
            self._task.cancel()
        if self._socket is not None:
            self._socket.close()
        if self._context is not None:
            self._context.destroy()
        self.loop.close()
        self.closed = True



"""
Useful callbacks for the Run Engine
"""

import logging
import os
import time as ttime
import warnings
from collections import OrderedDict, deque, namedtuple
from datetime import datetime
from enum import Enum
from functools import partial as _partial
from functools import wraps as _wraps
from itertools import count

from event_model import DocumentRouter

from ..utils import ensure_uid

logger = logging.getLogger(__name__)


def make_callback_safe(func=None, *, logger=None):
    """
    If the wrapped func raises any exceptions, log them but continue.

    This is intended to ensure that any failures in non-critical callbacks do
    not interrupt data acquisition. It should *not* be applied to any
    critical callbacks, such as ones that perform data-saving, but is well
    suited to callbacks that perform non-critical streaming visualization or
    data processing.

    To debug the issue causing a failure, it can be convenient to turn this
    off and let the failures raise. To do this, set the environment variable
    ``BLUESKY_DEBUG_CALLBACKS=1``.

    Parameters
    ----------
    func: callable
    logger: logging.Logger, optional

    Examples
    --------

    Decorate a callback to make sure it will not interrupt data acquisition if
    it fails.

    >>> @make_callback_safe
    ... def callback(name, doc):
    ...     ...
    """
    if func is None:
        return _partial(make_callback_safe, logger=logger)

    @_wraps(func)
    def inner(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except Exception:
            debug_mode = os.environ.get("BLUESKY_DEBUG_CALLBACKS", False)
            if logger is not None:
                if debug_mode:
                    msg = f"Exception in {func}"
                else:
                    msg = (
                        f"An exception raised in the callback {func} "
                        "is being suppressed to not interrupt plan "
                        "execution.  To investigate try setting the "
                        "BLUESKY_DEBUG_CALLBACKS env to '1'"
                    )
                logger.exception(msg)
            if debug_mode:
                raise

    return inner


def make_class_safe(cls=None, *, to_wrap=None, logger=None):
    """
    If the wrapped func raises any exceptions, log them but continue.

    This is intended to ensure that any failures in non-critical callbacks do
    not interrupt data acquisition. It should *not* be applied to any
    critical callbacks, such as ones that perform data-saving, but is well
    suited to callbacks that perform non-critical streaming visualization or
    data processing.

    To debug the issue causing a failure, it can be convenient to turn this
    off and let the failures raise. To do this, set the environment variable
    ``BLUESKY_DEBUG_CALLBACKS=1``.

    Parameters
    ----------
    cls: callable
    to_wrap: List[String], optional
        Names of methods of cls to wrap. Default is ``['__call__']``.
    logger: logging.Logger, optional

    Examples
    --------

    Decorate a class to make sure it will not interrupt data acquisition if
    it fails.

    >>> @make_class_safe
    ... class Callback(event_model.DocumentRouter):
    ...     ...
    """

    if cls is None:
        return _partial(make_class_safe, to_wrap=to_wrap, logger=logger)

    if to_wrap is None:
        to_wrap = ["__call__"]

    for f_name in to_wrap:
        setattr(cls, f_name, make_callback_safe(getattr(cls, f_name), logger=logger))

    return cls


class CallbackBase(DocumentRouter):
    log = None


class CallbackCounter:
    "As simple as it sounds: count how many times a callback is called."

    # Wrap itertools.count in something we can use as a callback.
    def __init__(self):
        self.counter = count()
        self(None, {})  # Pass a fake doc to prime the counter (start at 1).

    def __call__(self, name, doc):
        self.value = next(self.counter)


def print_metadata(name, doc):
    "Print all fields except uid and time."
    for field, value in sorted(doc.items()):
        # uid is returned by the RunEngine, and time is self-evident
        if field not in ["time", "uid"]:
            print(f"{field}: {value}")


def collector(field, output):
    """
    Build a function that appends data to a list.

    This is useful for testing but not advised for general use. (There is
    probably a better way to do whatever you want to do!)

    Parameters
    ----------
    field : str
        the name of a data field in an Event
    output : mutable iterable
        such as a list

    Returns
    -------
    func : function
        expects one argument, an Event dictionary
    """

    def f(name, event):
        output.append(event["data"][field])

    return f


def format_num(x, max_len=11, pre=5, post=5):
    if (abs(x) > 10**pre or abs(x) < 10**-post) and x != 0:
        x = f"%.{post}e" % x
    else:
        x = f"%{pre}.{post}f" % x

    return x


def get_obj_fields(fields):
    """
    If fields includes any objects, get their field names using obj.describe()

    ['det1', det_obj] -> ['det1, 'det_obj_field1, 'det_obj_field2']"
    """
    string_fields = []
    for field in fields:
        if isinstance(field, str):
            string_fields.append(field)
        else:
            try:
                field_list = sorted(field.describe().keys())
            except AttributeError:
                raise ValueError(  # noqa: B904
                    "Fields must be strings or objects with a 'describe' method that return a dict."
                )
            string_fields.extend(field_list)
    return string_fields


class CollectThenCompute(CallbackBase):
    def __init__(self):
        self._start_doc = None
        self._stop_doc = None
        self._events = deque()
        self._descriptors = deque()

    def start(self, doc):
        self._start_doc = doc
        super().start(doc)

    def descriptor(self, doc):
        self._descriptors.append(doc)
        super().descriptor(doc)

    def event(self, doc):
        self._events.append(doc)
        super().event(doc)

    def stop(self, doc):
        self._stop_doc = doc
        self.compute()
        super().stop(doc)

    def reset(self):
        self._start_doc = None
        self._stop_doc = None
        self._events.clear()
        self._descriptors.clear()

    def compute(self):
        raise NotImplementedError("This method must be defined by a subclass.")


@make_class_safe(logger=logger)
class LiveTable(CallbackBase):
    """Live updating table

    Parameters
    ----------
    fields : list
         List of fields to add to the table.

    stream_name : str, optional
         The event stream to watch for

    print_header_interval : int, optional
         Reprint the header every this many lines, defaults to 50

    min_width : int, optional
         The minimum width is spaces of the data columns.  Defaults to 12

    default_prec : int, optional
         Precision to use if it can not be found in descriptor, defaults to 3

    extra_pad : int, optional
         Number of extra spaces to put around the printed data, defaults to 1

    separator_lines : bool, optional
        Add empty lines before and after the printed table, default True

    logbook : callable, optional
        Must take a sting as the first positional argument

           def logbook(input_str):
                pass

    out : callable, optional
        Function to call to 'print' a line.  Defaults to `print`
    """

    _FMTLOOKUP = {
        "s": "{pad}{{{k}: >{width}.{prec}{dtype}}}{pad}",
        "f": "{pad}{{{k}: >{width}.{prec}{dtype}}}{pad}",
        "g": "{pad}{{{k}: >{width}.{prec}{dtype}}}{pad}",
        "d": "{pad}{{{k}: >{width}{dtype}}}{pad}",
    }
    _FMT_MAP = {
        "number": "f",
        "integer": "d",
        "string": "s",
        "boolean": "s",
    }
    _fm_sty = namedtuple("fm_sty", ["width", "prec", "dtype"])  # type: ignore
    water_mark = "{st[plan_type]} {st[plan_name]} ['{st[uid]:.8s}'] (scan num: {st[scan_id]})"
    ev_time_key = "SUPERLONG_EV_TIMEKEY_THAT_I_REALLY_HOPE_NEVER_CLASHES"

    def __init__(
        self,
        fields,
        *,
        stream_name="primary",
        print_header_interval=50,
        min_width=12,
        default_prec=3,
        extra_pad=1,
        separator_lines=True,
        logbook=None,
        out=print,
    ):
        super().__init__()
        self._header_interval = print_header_interval
        # expand objects
        self._fields = get_obj_fields(fields)
        self._stream = stream_name
        self._start = None
        self._stop = None
        self._descriptors = set()
        self._pad_len = extra_pad
        self._extra_pad = " " * extra_pad
        self._min_width = min_width
        self._default_prec = default_prec
        self._separator_lines = separator_lines
        self._format_info = OrderedDict(
            [
                ("seq_num", self._fm_sty(10 + self._pad_len, "", "d")),
                (self.ev_time_key, self._fm_sty(10 + 2 * extra_pad, 10, "s")),
            ]
        )
        self._rows = []
        self.logbook = logbook
        self._sep_format = None
        self._out = out

    def descriptor(self, doc):
        def patch_up_precision(p):
            try:
                return int(p)
            except (TypeError, ValueError):
                return self._default_prec

        if doc["name"] != self._stream:
            return

        self._descriptors.add(doc["uid"])

        dk = doc["data_keys"]
        for k in self._fields:
            width = max(self._min_width, len(k) + 2, self._default_prec + 1 + 2 * self._pad_len)
            try:
                dk_entry = dk[k]
            except KeyError:
                # this descriptor does not know about this key
                continue

            if dk_entry["dtype"] not in self._FMT_MAP:
                warnings.warn(  # noqa: B028
                    "The key {} will be skipped because LiveTable "
                    "does not know how to display the dtype {}"
                    "".format(k, dk_entry["dtype"])
                )
                continue

            if dk_entry["dtype"] == "boolean":
                prec = 5  # 5 is the length of the string "False"
            elif dk_entry["dtype"] == "string":
                # If string, set precision to length of the key minus padding
                prec = width - 1 - 2 * self._pad_len
            else:
                prec = patch_up_precision(dk_entry.get("precision", self._default_prec))

            fmt = self._fm_sty(width=width, prec=prec, dtype=self._FMT_MAP[dk_entry["dtype"]])

            self._format_info[k] = fmt

        self._sep_format = "+" + "+".join("-" * f.width for f in self._format_info.values()) + "+"
        self._main_fmnt = "|".join(
            "{{: >{w}}}{pad}".format(w=f.width - self._pad_len, pad=" " * self._pad_len)
            for f in self._format_info.values()
        )
        headings = [k if k != self.ev_time_key else "time" for k in self._format_info]
        self._header = "|" + self._main_fmnt.format(*headings) + "|"
        self._data_formats = OrderedDict(
            (
                k,
                self._FMTLOOKUP[f.dtype].format(
                    k=f"h{str(hash(k))}",
                    width=f.width - 2 * self._pad_len,
                    prec=f.prec,
                    dtype=f.dtype,
                    pad=self._extra_pad,
                ),
            )
            for k, f in self._format_info.items()
        )

        self._count = 0

        if self._separator_lines:
            self._print("\n")
        self._print(self._sep_format)
        self._print(self._header)
        self._print(self._sep_format)
        super().descriptor(doc)

    def event(self, doc):
        try:
            if ensure_uid(doc["descriptor"]) not in self._descriptors:
                return
            # shallow copy so we can mutate
            data = dict(doc["data"])
            self._count += 1
            if not self._count % self._header_interval:
                self._print(self._sep_format)
                self._print(self._header)
                self._print(self._sep_format)
            fmt_time = str(datetime.fromtimestamp(doc["time"]).time())
            data[self.ev_time_key] = fmt_time
            data["seq_num"] = doc["seq_num"]
            cols = [
                # If we have a bool, just `str` it.
                f.format(**{f"h{str(hash(k))}": str(data[k])})
                if ((k in data) and isinstance(data[k], (bool)))
                # If we have an enum, format with it's value
                else f.format(**{f"h{str(hash(k))}": data[k].value})
                if ((k in data) and isinstance(data[k], (Enum)))
                else f.format(**{f"h{str(hash(k))}": data[k]})
                # Show data[k] if k exists in this Event and is 'filled'.
                # (The latter is only applicable if the data is
                # externally-stored -- hence the fallback to `True`.)
                if ((k in data) and doc.get("filled", {}).get(k, True))
                # Otherwise use a placeholder of whitespace.
                else " " * self._format_info[k].width
                for k, f in self._data_formats.items()
            ]
            self._print("|" + "|".join(cols) + "|")
        except Exception as ex:
            if self.log is not None:
                self.log.exception(ex)
            self._print(f"{{k:*^{self._min_width}}}".format(k=" failed to format row "))
            if os.environ.get("BLUESKY_DEBUG_CALLBACKS", False):
                raise ex
        super().event(doc)

    def stop(self, doc):
        if ensure_uid(doc["run_start"]) != self._start["uid"]:
            return

        # This sleep is just cosmetic. It improves the odds that the bottom
        # border is not printed until all the rows from events are printed,
        # avoiding this ugly scenario:
        #
        # |         4 | 22:08:56.7 |      0.000 |
        # +-----------+------------+------------+
        # generator scan ['6d3f71'] (scan num: 1)
        # Out[2]: |         5 | 22:08:56.8 |      0.000 |
        ttime.sleep(0.1)

        if self._sep_format is not None:
            self._print(self._sep_format)
        self._stop = doc

        wm = self.water_mark.format(st=self._start)
        self._out(wm)
        if self.logbook:
            self.logbook("\n".join([wm] + self._rows))
        if self._separator_lines:
            self._print("\n")
        super().stop(doc)

    def start(self, doc):
        self._rows = []
        self._start = doc
        self._stop = None
        self._sep_format = None
        super().start(doc)

    def _print(self, out_str):
        self._rows.append(out_str)
        self._out(out_str)



import asyncio
import os

import numpy as np
import packaging
import pytest

from bluesky.run_engine import RunEngine, TransitionError


@pytest.fixture(scope="function", params=[False, True])
def RE(request):
    loop = asyncio.new_event_loop()
    loop.set_debug(True)
    RE = RunEngine({}, call_returns_result=request.param, loop=loop)

    def clean_event_loop():
        if RE.state not in ("idle", "panicked"):
            try:
                RE.halt()
            except TransitionError:
                pass
        loop.call_soon_threadsafe(loop.stop)
        RE._th.join()
        loop.close()

    request.addfinalizer(clean_event_loop)
    return RE


@pytest.fixture(scope="function")
def hw(tmp_path):
    import ophyd
    from ophyd.sim import hw

    # ophyd 1.4.0 added support for customizing the directory used by simulated
    # hardware that generates files
    if packaging.version.Version(ophyd.__version__) >= packaging.version.Version("1.4.0"):
        return hw(str(tmp_path))
    else:
        return hw()


# vendored from ophyd.sim
class NumpySeqHandler:
    specs = {"NPY_SEQ"}

    def __init__(self, filename, root=""):
        self._name = os.path.join(root, filename)

    def __call__(self, index):
        return np.load(f"{self._name}_{index}.npy")

    def get_file_list(self, datum_kwarg_gen):
        "This method is optional. It is not needed for access, but for export."
        return ["{name}_{index}.npy".format(name=self._name, **kwargs) for kwargs in datum_kwarg_gen]


@pytest.fixture(scope="function")
def db(request):
    """Return a data broker"""
    try:
        from databroker import temp

        db = temp()
        return db
    except ImportError:
        pytest.skip("Databroker v2 still missing temp.")
    except ValueError:
        pytest.skip("Intake is failing for unknown reasons.")


@pytest.fixture(autouse=True)
def cleanup_any_figures(request):
    import matplotlib.pyplot as plt

    "Close any matplotlib figures that were opened during a test."
    plt.close("all")
